from ophyd import (
    PseudoPositioner,
    EpicsMotor,
    Component as Cpt,
    FormattedComponent as FCpt,
    PseudoSingle,
    PVPositioner,
    EpicsSignal,
    EpicsSignalRO,
)
from ophyd.ophydobj import OphydObject
from ophyd.pseudopos import pseudo_position_argument, real_position_argument


from .._iconfig import load_config
from .instrument_registry import registry
from .monochromator import Monochromator, IDTracking


__all__ = ["EnergyPositioner", "load_energy_positioner"]


class Undulator(PVPositioner):
    setpoint = Cpt(EpicsSignal, ":ScanEnergy.VAL")
    readback = Cpt(EpicsSignalRO, ":Energy.VAL")
    done = Cpt(EpicsSignalRO, ":Busy.VAL", kind="omitted")
    stop_signal = Cpt(EpicsSignal, ":Stop.VAL", kind="omitted")


class EnergyPositioner(PseudoPositioner):
    """The operational energy of the beamline.

    Responsible for setting both mono and ID energy with an optional
    ID offset. Setting the *energy* component will propagate to both
    real devices and so ``EnergyPositioner().energy`` is a good
    candidate for a positioner for Bluesky plans.

    Currently, the offset between the insertion device and the
    monochromator is fixed. In the future this will be replaced with a
    more sophisticated calculation.

    .. todo::

       Insert functionality to have a non-constant ID offset.

    Attributes
    ==========

    id_offset
      The offset for the insertion device relative to the mono energy.
    energy
      The pseudo positioner for the forward calculation.
    mono_energy
      The real component for the monochromator energy.
    id_energy
      The real component for the insertion device energy.

    Parameters
    ==========
    mono_pv
      The process variable for the monochromator energy PV.
    id_prefix
      The prefix for the insertion device energy, such that
      f"{id_prefix}:Energy.VAL" reaches the energy readback value.
    id_offset_pv
      The PV address for the offset between monochromator and
      insertion device.
    id_tracking_pv
      The PV address for whether the ID gets tracked automatically in
      EPICS.

    """

    # Pseudo axes
    energy: OphydObject = Cpt(PseudoSingle, kind="hinted")

    # Equivalent real axes
    mono_energy: OphydObject = FCpt(EpicsMotor, "{mono_pv}", kind="normal")
    # id_offset: float = 300.  # In eV
    id_tracking: OphydObject = FCpt(EpicsSignal, "{id_tracking_pv}", kind="config")
    id_offset: OphydObject = FCpt(EpicsSignal, "{id_offset_pv}", kind="config")
    id_energy: OphydObject = FCpt(Undulator, "{id_prefix}", kind="normal")

    def __init__(
        self,
        mono_pv: str,
        id_prefix: str,
        id_offset_pv: str,
        id_tracking_pv: str,
        *args,
        **kwargs,
    ):
        """INIT DOCSTRING"""
        self.mono_pv = mono_pv
        self.id_offset_pv = id_offset_pv
        self.id_prefix = id_prefix
        self.id_tracking_pv = id_tracking_pv
        super().__init__(*args, **kwargs)
        self.stage_sigs[self.id_tracking] = IDTracking.OFF

    @pseudo_position_argument
    def forward(self, target_energy):
        "Given a target energy, transform to the mono and ID energies."
        id_offset_ev = self.id_offset.get(use_monitor=True)
        return self.RealPosition(
            mono_energy=target_energy.energy,
            id_energy=(target_energy.energy + id_offset_ev) / 1000.0,
        )

    @real_position_argument
    def inverse(self, device_energy):
        "Given a position in mono and ID energy, transform to the target energy."
        return self.PseudoPosition(
            energy=device_energy.mono_energy,
        )


def load_energy_positioner(config=None):
    # Load PV's from config
    if config is None:
        config = load_config()
    mono_suffix = Monochromator.energy.suffix
    id_offset_suffix = Monochromator.id_offset.suffix
    id_tracking_suffix = Monochromator.id_tracking.suffix
    mono_prefix = config["monochromator"]["ioc"]
    id_prefix = config["undulator"]["ioc"]
    # Create energy positioner
    energy_positioner = EnergyPositioner(
        name="energy",
        mono_pv=f"{mono_prefix}{mono_suffix}",
        id_offset_pv=f"{mono_prefix}{id_offset_suffix}",
        id_tracking_pv=f"{mono_prefix}{id_tracking_suffix}",
        id_prefix=f"{id_prefix}",
    )
    registry.register(energy_positioner)



import pyqtgraph
import numpy as np
import pydm
from unittest import mock

from firefly.main_window import FireflyMainWindow
from firefly.area_detector_viewer import AreaDetectorViewerDisplay
from haven.instrument.camera import load_cameras


def test_open_camera_viewer_actions(ffapp, qtbot, sim_camera):
    # Now get the cameras ready
    ffapp.prepare_camera_windows()
    assert hasattr(ffapp, "camera_actions")
    assert len(ffapp.camera_actions) == 1
    # Launch an action and see that a window opens
    list(ffapp.camera_actions.values())[0].trigger()
    assert "FireflyMainWindow_camera_s255id-gige-A" in ffapp.windows.keys()


def test_open_area_detector_viewer_actions(ffapp, qtbot, sim_camera):
    # Get the area detector parts ready
    ffapp.prepare_area_detector_windows()
    assert hasattr(ffapp, "area_detector_actions")
    assert len(ffapp.area_detector_actions) == 1
    # Launch an action and see that a window opens
    list(ffapp.area_detector_actions.values())[0].trigger()
    assert "FireflyMainWindow_area_detector_s255id-gige-A" in ffapp.windows.keys()


def test_image_plotting(ffapp, qtbot, sim_camera):
    FireflyMainWindow()
    display = AreaDetectorViewerDisplay(macros={"AD": sim_camera.name})
    assert isinstance(display.image_view, pyqtgraph.ImageView)
    assert isinstance(display.image_channel, pydm.PyDMChannel)
    # Give it some grayscale data
    source_img = np.random.randint(0, 256, size=(54, 64))
    # For some reason, PyDMConnection mis-shapes the data
    reshape_img = np.reshape(source_img, source_img.shape[::-1])
    display.image_channel.value_slot(reshape_img)
    # See if we get a plottable image out
    new_image = display.image_view.getImageItem()
    assert new_image.image is not None
    assert new_image.image.shape == (54, 64)
    np.testing.assert_equal(new_image.image, source_img)
    # Give it some RGB data
    source_img = np.random.randint(0, 256, size=(54, 64, 3))
    # For some reason, PyDMConnection mis-shapes the data
    reshape_img = np.reshape(source_img, source_img.shape[::-1])
    display.image_channel.value_slot(reshape_img)
    # See if we get a plottable image out
    new_image = display.image_view.getImageItem()
    assert new_image.image.shape == (54, 64, 3)
    display.image_channel.disconnect()


def test_caqtdm_window(ffapp, sim_camera):
    FireflyMainWindow()
    display = AreaDetectorViewerDisplay(macros={"AD": sim_camera.name})
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    display._open_caqtdm_subprocess.assert_called_once_with(
        f"start_{sim_camera.prefix}_caqtdm"
    )



"""A bluesky plan to scan the X-ray energy over an X-ray edge and
capture detector signals.

"""

from typing import Union, Sequence, Optional, Mapping
from collections import ChainMap
import warnings
import logging

import numpy as np

from ..energy_ranges import ERange, KRange, merge_ranges
from .. import exceptions
from .energy_scan import energy_scan
from ..typing import DetectorList
from ..preprocessors import baseline_decorator, shutter_suspend_decorator

log = logging.getLogger(__name__)


log = logging.getLogger(__name__)


__all__ = ["xafs_scan"]


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]  # noqa: E203


# @shutter_suspend_decorator()
@baseline_decorator()
def xafs_scan(
    E_min: float,
    *E_params: Sequence[float],
    k_step: Optional[float] = None,
    k_exposure: Optional[float] = None,
    k_max: Optional[float] = None,
    k_weight: float = 0.0,
    E0: Union[float, str] = 0,
    detectors: DetectorList = "ion_chambers",
    energy_positioners: Sequence = ["energy"],
    time_positioners: Sequence = ["I0_exposure_time"],
    md: Mapping = {},
):
    """Collect a spectrum by scanning X-ray energies relative to an
    absorbance edge.

    Used like:

    .. code-block:: python

        xafs_scan(energy, step, exposure, energy, step, exposure energy, ...)

    The optional parameter *E0* can either be an absolute energy in
    electron-volts, or a string of the form "Ni_K" to be looked up in
    a reference table. If omitted, energies will be absolute and
    K-space parameters should be omitted.

    If *E0* is given, then energies should then be given relative to
    this edge in groups of ``(min, step, exposure, max)``. More than
    one range can be specified by giving additional groups of ``(step,
    exposure, max)`` since the maximum energy of the previous range
    will be the minimum energy of the subsequent range.

    For example, the following invocation will scan three regions
    relative to the Ni K-edge:

    - Pre-edge: from 100 eV below the edge to 30 eV below the edge in
      2 eV steps with 0.5 sec exposure.
    - Edge: from 30 eV below the edge to 50 eV above the edge in 0.1
      eV steps with 1 sec exposure.
    - EXAFS: from 50 eV above the edge to K=8 Å⁻ in 0.2 Å⁻ steps with
      1.5 sec exposure.

    .. code-block:: python

        xafs_scan(-100, 2, 0.5, -30, 0.1, 1., 50,
                  k_step=0.2, k_exposure=1.5, k_max=8,
                  E0="Ni_K")

    For measuring the extended structure (EXAFS) *k_step*, *k_max*,
    and *k_exposure* can be given instead of the equivalent
    energies. *k_weights* will apply longer exposure times to higher K
    values.

    *detectors*, *energy_positioners*, and *time_positioners* are
     mostly useful for debugging.

    Parameters
    ==========
    E_min
      The starting energy for the first energy range, in eV.
    E_params
      Should be any number of parameters of the form ``energy_step,
      exposure, energy, energy_step, exposure, energy, ...``. Energies
      should be in eV and *exposure* in seconds.
    k_step
      Wavenumber (k) step for the EXAFS region, in Å⁻.
    k_exposure
      Base exposure time for the EXAFS region, in seconds.
    k_max
      Last energy for the EXAFS region, in eV.
    k_weight
      Weighting factor for exposure time at higher energies in the
      EXAFS region. Default (0) results in constant exposure.
    E0
      An edge energy, in eV, or name of the edge of the form
      "Ni_L3". All energies will be relative to this value.
    detectors
      Overrides the list of Ophyd signals to record during this
      scan. Useful for testing.
    energy_positioners
      Overrides the list of Ophyd positioners used to set energy
      during this scan. Useful for testing.
    time_positioners
      Overrides the list of Ophyd positioners used to set exposure
      time during this scan. Useful for testing.

    """
    # Make sure the right number of energies have been given
    # Turn the energies in energy ranges
    curr_E_min = E_min
    energy_ranges = []
    for E_step, E_exposure, E_max in chunks(E_params, 3):
        energy_ranges.append(ERange(curr_E_min, E_max, E_step, exposure=E_exposure))
        curr_E_min = E_max
    # Check for k-space region but no E0 specified
    has_krange = not any([k is None for k in [k_step, k_exposure, k_max]])
    if has_krange and E0 == 0:
        msg = "Specified K-range for scan, but no *E0* given."
        warnings.warn(msg)
        log.warning(msg)
    if has_krange:
        energy_ranges.append(
            KRange(
                E_min=curr_E_min,
                k_max=k_max,
                k_step=k_step,
                k_weight=k_weight,
                exposure=k_exposure,
            )
        )
    # Convert energy ranges to energy list and exposure list
    energies, exposures = merge_ranges(*energy_ranges)
    if len(energies) < 1:
        raise exceptions.NoEnergies("Plan would not produce any energy points.")
    # Execute the energy scan
    yield from energy_scan(
        energies=energies,
        exposure=exposures,
        E0=E0,
        detectors=detectors,
        energy_positioners=energy_positioners,
        time_positioners=time_positioners,
        md=ChainMap(md, {"plan_name": "xafs_scan"}),
    )



from dataclasses import dataclass

import numpy as np
import pint
from scipy import constants


__all__ = ["ERange", "KRange"]


DEFAULT_EXPOSURE = 0.5

# Calculate the conversion coefficient from k -> eV
ureg = pint.UnitRegistry()
hbar = constants.hbar * ureg.joule * ureg.second
c = constants.speed_of_light * ureg.meter / ureg.second
m_e = (
    constants.physical_constants["electron mass energy equivalent in MeV"][0]
    * 1e6
    * ureg.electron_volt
)
ALPHA = hbar**2 * c**2 / 2 / m_e
ALPHA = ALPHA.to("electron_volt * angstrom * angstrom").magnitude


@dataclass
class EnergyRange:
    """A range of energies used for scanning."""

    def energies(self):
        raise NotImplementedError

    def exposures(self):
        raise NotImplementedError


@dataclass
class ERange(EnergyRange):
    """A range of energies used for scanning.

    All values are assumed to be in electron-volts. If *E0* is a

    Parameters
    ==========

    E_min
      Starting energy of the range, in eV.
    E_max
      Ending energy of the range, in eV.
    E_step
      Step-size between energies, in eV.
    exposure
      How long to spend at each energy, in seconds.

    """

    E_min: float
    E_max: float
    E_step: float = 1.0
    exposure: float = DEFAULT_EXPOSURE

    def energies(self):
        """Convert the range to a sequence of actual energy values, in eV."""
        return np.arange(self.E_min, self.E_max + self.E_step, self.E_step)

    def exposures(self):
        """Convert the range to a sequence of exposure times, in seconds."""
        return [self.exposure] * len(self.energies())


@dataclass
class KRange(EnergyRange):
    """A range of energies used for scanning in the EXAFS region.

    Values are a mix of electron-volts and K-space values.

    Parameters
    ==========

    E_min
      Starting energy of the range, in eV.
    k_max
      Ending energy of the range, in Å⁻.
    k_step
      Step-size between energies, in Å⁻.
    k_weight
      Weighting factor for longer exposures at higher energies.
    exposure
      How long to spend at each energy, in seconds.

    """

    E_min: float
    k_max: float
    k_step: float = 0.1
    k_weight: float = 0.0
    exposure: float = DEFAULT_EXPOSURE

    def energies(self):
        """Calculates photon energies in units of eV."""
        return self.wavenumber_to_energy(self.wavenumbers())

    def energy_to_wavenumber(self, energy):
        return np.sqrt(energy / ALPHA)

    def wavenumber_to_energy(self, wavenumber):
        return wavenumber**2 * ALPHA

    def wavenumbers(self):
        """Calculates wavenumbers (k) for the photo-electron in units Å⁻."""
        k_min = self.energy_to_wavenumber(self.E_min)
        ks = np.arange(k_min, self.k_max + self.k_step, self.k_step)
        return ks

    def exposures(self):
        ks = self.wavenumbers()
        return self.exposure * (ks / np.min(ks)) ** self.k_weight


def merge_ranges(*ranges, default_exposure=DEFAULT_EXPOSURE):
    """Combine multiple energy ranges.

    If any of *ranges* is a instance of ``EnergyRange`` or one of its
    subclasses, then the object will be decomposed to it's constituent
    energies and exposure times. Otherwise, the value will be used
    as-is for the list of energies, and *default_exposure* will be
    used for the exposure time.

    Results will be sorted by energy, and redundant energies will be
    removed with the exposure time being taken from the first instance
    of the particular energy encountered.

    Returns
    =======
    energies
      A sequence of energies, in eV.
    exposures
      A sequence of exposure times, in seconds.

    """
    energies = []
    exposures = []
    for rng in ranges:
        if not isinstance(rng, EnergyRange):
            # Not an energy range, assume it's a float or something
            energies.append(rng)
            exposures.append(default_exposure)
        else:
            # An energy range, so break it down
            energies.extend(rng.energies())
            exposures.extend(rng.exposures())
    # Convert to proper arrays, and remove duplicate energies
    energies, unique_idx = np.unique(
        np.asarray(energies, dtype=float), return_index=True
    )
    exposures = np.asarray(exposures, dtype=float)[unique_idx]
    return energies, exposures



from ophyd import (
    Device,
    Component,
    FormattedComponent,
    EpicsSignal,
    EpicsSignalRO,
)


__all__ = ["ScalerTriggered"]


class ScalerTriggered:
    """A mix-in for triggering a device using the scaler.

    The device does not have to be a channel on the scaler, enabling
    hardware triggering of other devices.

    If this is a used as part of a component, then triggering is left
    up to the parent device.

    If this is used as part of a top-level Device, then the device
    should have a *scaler_prefix* attribute with the PV prefix to the
    scaler (e.g. "25idcVME:scaler1"), though this is only used to
    coordinate statuses for multiple devices using the same scaler.

    """

    scaler_prefix = None
    _statuses = {}

    def trigger(self, *args, **kwargs):
        is_top_device = getattr(self, "parent", None) is None
        if is_top_device:
            # This is the top-level device, so trigger it
            return self._trigger_scaler(*args, **kwargs)
        else:
            # This is a sub-component of a device, so trigger the parent
            return self.parent.trigger()

    def _trigger_scaler(self, *args, **kwargs):
        # Figure out if there's already a trigger active
        previous_status = self._statuses.get(self.scaler_prefix)
        is_idle = previous_status is None or previous_status.done
        # Trigger the detector if not already running, and update the status dict
        if is_idle:
            new_status = super().trigger(*args, **kwargs)
            self._statuses[self.scaler_prefix] = new_status
        else:
            new_status = previous_status
        return new_status


class ScalerSignal(ScalerTriggered, EpicsSignal):
    ...


class ScalerSignalRO(ScalerTriggered, EpicsSignalRO):
    ...



from enum import IntEnum
from apstools.devices import CamMixin_V34, SingleTrigger_V34
from ophyd import (
    ADComponent as ADCpt,
    DetectorBase as OphydDetectorBase,
    SimDetectorCam,
    Lambda750kCam,
    EigerDetectorCam,
    SingleTrigger,
    Kind,
)
from ophyd.areadetector.filestore_mixins import FileStoreHDF5IterativeWrite
from ophyd.areadetector.plugins import (
    HDF5Plugin_V34,
    HDF5Plugin_V31,
    ImagePlugin_V34,
    ImagePlugin_V31,
    PvaPlugin_V34,
    PvaPlugin_V31,
    TIFFPlugin_V34,
    TIFFPlugin_V31,
    ROIPlugin_V34,
    ROIPlugin_V31,
    StatsPlugin_V31 as OphydStatsPlugin_V31,
    StatsPlugin_V34 as OphydStatsPlugin_V34,
    OverlayPlugin,
)


from .._iconfig import load_config
from .instrument_registry import registry
from .. import exceptions


__all__ = ["Eiger500K", "Lambda250K", "SimDetector"]


class SimDetectorCam_V34(CamMixin_V34, SimDetectorCam):
    ...


class WriteModes(IntEnum):
    SINGLE = 0
    CAPTURE = 1
    STREAM = 2


class Capture(IntEnum):
    STOP = 0
    START = 1


class StageCapture:
    """Mixin to prepare NDPlugin file capture mode.

    Sets the number of captures to zero (infinite), and starts
    capturing. Then when the device gets unstaged, capturing turns
    back off.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Stage the capture button as well
        self.stage_sigs[self.file_write_mode] = WriteModes.STREAM
        self.stage_sigs[self.capture] = Capture.START
        self.stage_sigs[self.num_capture] = 0


class MyHDF5Plugin(FileStoreHDF5IterativeWrite, HDF5Plugin_V34):
    """
    Add data acquisition methods to HDF5Plugin.
    * ``stage()`` - prepare device PVs befor data acquisition
    * ``unstage()`` - restore device PVs after data acquisition
    * ``generate_datum()`` - coordinate image storage metadata
    """

    def stage(self):
        self.stage_sigs.move_to_end("capture", last=True)
        super().stage()


class DetectorBase(OphydDetectorBase):
    def __init__(self, *args, description=None, **kwargs):
        super().__init__(*args, **kwargs)
        if description is None:
            description = self.name
        self.description = description

    overlays = ADCpt(OverlayPlugin, "Over1:")


class StatsMixin:
    _default_read_attrs = [
        "max_value",
        "min_value",
        "min_xy.x",
        "max_xy.x",
        "min_xy.y",
        "max_xy.y",
        "total",
        "net",
        "mean_value",
        "sigma_value",
    ]


class StatsPlugin_V31(StatsMixin, OphydStatsPlugin_V31):
    ...


class StatsPlugin_V34(StatsMixin, OphydStatsPlugin_V34):
    ...


class SimDetector(SingleTrigger_V34, DetectorBase):
    """
    ADSimDetector
    SingleTrigger:
    * stop any current acquisition
    * sets image_mode to 'Multiple'
    """

    cam = ADCpt(SimDetectorCam_V34, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    hdf1 = ADCpt(
        type("HDF5Plugin", (StageCapture, HDF5Plugin_V34), {}),
        "HDF1:",
        # write_path_template="/tmp/",
        # read_path_template=READ_PATH_TEMPLATE,
    )
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)


class TIFFPlugin(StageCapture, TIFFPlugin_V31):
    _default_read_attrs = ["full_file_name"]


class HDF5Plugin(StageCapture, HDF5Plugin_V31):
    _default_read_attrs = ["full_file_name"]


class Lambda250K(SingleTrigger, DetectorBase):
    """
    A Lambda 250K area detector device.
    """

    cam = ADCpt(Lambda750kCam, "cam1:")
    image = ADCpt(ImagePlugin_V31, "image1:")
    pva = ADCpt(PvaPlugin_V31, "Pva1:")
    tiff = ADCpt(TIFFPlugin, "TIFF1:", kind=Kind.normal)
    hdf1 = ADCpt(HDF5Plugin, "HDF1:", kind=Kind.normal)
    roi1 = ADCpt(ROIPlugin_V31, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V31, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V31, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V31, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V31, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V31, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V31, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V31, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V31, "Stats5:", kind=Kind.normal)

    _default_read_attrs = [
        "stats1",
        "stats2",
        "stats3",
        "stats4",
        "stats5",
        "hdf1",
        "tiff",
    ]


class Eiger500K(SingleTrigger, DetectorBase):
    """
    A Eiger S 500K area detector device.
    """

    cam = ADCpt(EigerDetectorCam, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    tiff = ADCpt(TIFFPlugin, "TIFF1:", kind=Kind.normal)
    hdf1 = ADCpt(HDF5Plugin, "HDF1:", kind=Kind.normal)
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)

    _default_read_attrs = [
        "stats1",
        "stats2",
        "stats3",
        "stats4",
        "stats5",
        "hdf1",
        "tiff",
    ]


def load_area_detectors(config=None):
    if config is None:
        config = load_config()
    # Create the area detectors defined in the configuration
    for name, adconfig in config.get("area_detector", {}).items():
        DeviceClass = globals().get(adconfig["device_class"])
        # Check that it's a valid device class
        if DeviceClass is None:
            msg = f"area_detector.{name}.device_class={adconfig['device_class']}"
            raise exceptions.UnknownDeviceConfiguration(msg)
        # Create the device
        det = DeviceClass(
            prefix=f"{adconfig['prefix']}:",
            name=name,
            labels={"area_detectors"},
        )
        registry.register(det)



from haven.instrument.xray_source import load_xray_sources


def test_load_xray_sources(sim_registry):
    load_xray_sources()
    # Check that the undulator was added to the registry
    dev = sim_registry.find(label="xray_sources")
    assert dev.prefix == "ID255:"
    assert dev.gap.pvname == "ID255:Gap"



"""An assortment of tests for plans defined in Haven.

NOTE: This file uses the older unittest framework. We have since moved
to pytest, and new tests should be added to a plan-specific test file.

"""

import unittest
import warnings

import numpy as np
from ophyd import sim
import pytest

from run_engine import RunEngineStub
from haven import align_slits, energy_scan, xafs_scan, registry, KRange


class PlanUnitTests(unittest.TestCase):
    """Unit tests for the beamline-specific plans.

    These tests do not require connected IOC's and are expected not to
    modify actual motor PVs, etc. Integration tests are run
    else-where.

    """

    mono_motor = sim.SynAxis(name="mono_energy", labels={"motors", "energies"})
    exposure_motor = sim.Signal(name="exposure")
    id_gap_motor = sim.SynAxis(name="id_gap_energy", labels={"motors", "energies"})
    RE = RunEngineStub(call_returns_result=True)

    def setUp(self):
        registry.clear()
        # Register an ion chamber
        I0 = sim.SynGauss(
            "I0",
            sim.motor,
            "motor",
            center=-0.5,
            Imax=1,
            sigma=1,
            labels={"ion_chambers"},
        )
        # Register the energy positioner
        exposure_time = sim.SynAxis(name="I0_exposure_time")
        energy = sim.SynAxis(name="energy")
        for dev in [I0, exposure_time, energy]:
            registry.register(dev)


class AlignSlitsTests(PlanUnitTests):
    def test_align_slits(self):
        """Check the plan to aligns the slits."""
        # Prepare a fake detector and slit motor
        slit_motor = sim.motor
        I0 = sim.SynGauss(
            "det",
            sim.motor,
            "motor",
            center=-0.5,
            Imax=1,
            sigma=1,
            labels={"detectors"},
        )
        # Execute the plan
        self.RE(align_slits(slit_motors=[slit_motor], ion_chamber=I0))
        # Check that the slit positions have been set
        assert slit_motor.position == pytest.approx(-0.5)

    def test_warn_poor_fit(self):
        """Check that the plan emits a warning when no good fit is detected."""
        # Prepare a fake detector and slit motor
        slit_motor = sim.motor
        I0 = sim.SynGauss(
            "det",
            sim.motor,
            "motor",
            center=-0.5,
            noise="poisson",
            noise_multiplier=100,
            Imax=1,
            sigma=1,
            labels={"detectors"},
        )
        # Execute the plan, catching warnings
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            self.RE(align_slits(slit_motors=[slit_motor], ion_chamber=I0))
            # Check that a warning was raised
            self.assertEqual(len(w), 1)
            self.assertIn("Poor fit while centering motor", str(w[0].message))


class EnergyScanTests(PlanUnitTests):
    energies = np.linspace(8300, 8500, 100)
    exposure_time = 1e-3

    def test_energy_scan_basics(self):
        # Set up fake detectors and motors
        I0_exposure = sim.SynAxis(
            name="I0_exposure",
            labels={
                "exposures",
            },
        )
        It_exposure = sim.SynAxis(
            name="It_exposure",
            labels={
                "exposures",
            },
        )
        I0 = sim.SynGauss(
            name="I0",
            motor=self.mono_motor,
            motor_field="mono_energy",
            center=np.median(self.energies),
            Imax=1,
            sigma=1,
            labels={"detectors"},
        )
        It = sim.SynSignal(
            func=lambda: 1.0,
            name="It",
            exposure_time=self.exposure_time,
        )
        # Execute the plan
        scan = energy_scan(
            self.energies,
            detectors=[I0, It],
            exposure=self.exposure_time,
            energy_positioners=[self.mono_motor, self.id_gap_motor],
            time_positioners=[I0_exposure, It_exposure],
        )
        self.RE(scan)
        # Check that the mono and ID gap ended up in the right position
        self.assertEqual(self.mono_motor.get().readback, np.max(self.energies))
        self.assertEqual(self.id_gap_motor.get().readback, np.max(self.energies))
        self.assertEqual(I0_exposure.get().readback, self.exposure_time)
        self.assertEqual(It_exposure.get().readback, self.exposure_time)

    def test_raises_on_empty_positioners(self):
        with self.assertRaises(ValueError):
            self.RE(energy_scan(self.energies, energy_positioners=[]))


class XafsScanTests(PlanUnitTests):
    E0 = 10000  # in electron-volts

    def test_single_range(self):
        expected_energies = np.arange(9990, 10001, step=1)
        expected_exposures = np.asarray([1.0])
        scan = xafs_scan(
            -10,
            1,
            1,
            0,
            E0=self.E0,
            energy_positioners=[self.mono_motor],
            time_positioners=[self.exposure_motor],
        )
        # Check that the mono motor is moved to the correct positions
        scan_list = list(scan)
        real_energies = [
            i.args[0]
            for i in scan_list
            if i[0] == "set" and i.obj.name == "mono_energy"
        ]
        np.testing.assert_equal(real_energies, expected_energies)
        # Check that the exposure is set correctly
        real_exposures = [
            i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
        ]
        np.testing.assert_equal(real_exposures, expected_exposures)

    def test_multi_range(self):
        expected_energies = np.concatenate(
            [
                np.arange(9990, 10001, step=2),
                np.arange(10001, 10011, step=1),
            ]
        )
        expected_exposures = np.asarray([0.5, 1.0])
        scan = xafs_scan(
            -10,
            2,
            0.5,
            0,
            1,
            1.0,
            10,
            E0=self.E0,
            energy_positioners=[self.mono_motor],
            time_positioners=[self.exposure_motor],
        )
        # Check that the mono motor is moved to the correct positions
        scan_list = list(scan)
        real_energies = [
            i.args[0]
            for i in scan_list
            if i[0] == "set" and i.obj.name == "mono_energy"
        ]
        np.testing.assert_equal(real_energies, expected_energies)
        # Check that the exposure is set correctly
        real_exposures = [
            i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
        ]
        np.testing.assert_equal(real_exposures, expected_exposures)

    def test_exafs_k_range(self):
        """Ensure that passing in k_min, etc. produces an energy range
        in K-space.

        """
        krange = KRange(E_min=10, k_max=14, k_step=0.5, k_weight=0.5, exposure=0.75)
        expected_energies = krange.energies() + self.E0
        expected_exposures = krange.exposures()
        scan = xafs_scan(
            E_min=10,
            k_step=0.5,
            k_max=14,
            k_exposure=0.75,
            k_weight=0.5,
            E0=self.E0,
            energy_positioners=[self.mono_motor],
            time_positioners=[self.exposure_motor],
        )
        # Check that the mono motor is moved to the correct positions
        scan_list = list(scan)
        real_energies = [
            i.args[0]
            for i in scan_list
            if i[0] == "set" and i.obj.name == "mono_energy"
        ]
        np.testing.assert_equal(real_energies, expected_energies)
        # Check that the exposure is set correctly
        real_exposures = [
            i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
        ]
        np.testing.assert_equal(real_exposures, expected_exposures)

    def test_named_E0(self):
        expected_energies = np.concatenate(
            [
                np.arange(8323, 8334, step=2),
                np.arange(8334, 8344, step=1),
            ]
        )
        expected_exposures = np.asarray([0.5, 1.0])
        scan = xafs_scan(
            -10,
            2,
            0.5,
            0,
            1,
            1.0,
            10,
            E0="Ni_K",
            energy_positioners=[self.mono_motor],
            time_positioners=[self.exposure_motor],
        )
        # Check that the mono motor is moved to the correct positions
        scan_list = list(scan)
        real_energies = [
            i.args[0]
            for i in scan_list
            if i[0] == "set" and i.obj.name == "mono_energy"
        ]
        np.testing.assert_equal(real_energies, expected_energies)
        # Check that the exposure is set correctly
        real_exposures = [
            i.args[0] for i in scan_list if i[0] == "set" and i.obj.name == "exposure"
        ]
        np.testing.assert_equal(real_exposures, expected_exposures)

    def test_remove_duplicate_energies(self):
        plan = xafs_scan(
            -4,
            2,
            1.0,
            6,
            34,
            1.0,
            40,
            E0=8333,
            energy_positioners=[self.mono_motor],
            time_positioners=[self.exposure_motor],
        )
        msgs = list(plan)
        set_msgs = [
            m for m in msgs if m.command == "set" and m.obj.name == "mono_energy"
        ]
        read_msgs = [m for m in msgs if m.command == "read" and m.obj.name == "I0"]
        energies = [m.args[0] for m in set_msgs]
        # Make sure we only read each point once
        assert len(read_msgs) == len(energies)



import logging
from collections import OrderedDict
from pathlib import Path
from dataclasses import dataclass, field
from typing import Optional, Union, Mapping, Sequence
from functools import partial
import subprocess

from qtpy import QtWidgets, QtCore
from qtpy.QtWidgets import QAction
from qtpy.QtCore import Slot, QThread, Signal, QObject
import qtawesome as qta
import pydm
from pydm.application import PyDMApplication
from pydm.display import load_file
from pydm.utilities.stylesheet import apply_stylesheet
from bluesky_queueserver_api import BPlan
from bluesky_queueserver_api.zmq import REManagerAPI
import pyqtgraph as pg

from haven.exceptions import ComponentNotFound
from haven import HavenMotor, registry, load_config
import haven
from .main_window import FireflyMainWindow, PlanMainWindow
from .ophyd_plugin import OphydPlugin
from .queue_client import QueueClient, QueueClientThread

generator = type((x for x in []))

__all__ = ["ui_dir", "FireflyApplication"]


log = logging.getLogger(__name__)


ui_dir = Path(__file__).parent


pg.setConfigOption("background", (252, 252, 252))
pg.setConfigOption("foreground", (0, 0, 0))


class FireflyApplication(PyDMApplication):
    default_display = None
    xafs_scan_window = None

    # Actions for controlling the queueserver
    start_queue_action: QAction
    pause_runengine_action: QAction
    pause_runengine_now_action: QAction
    resume_runengine_action: QAction
    stop_runengine_action: QAction
    abort_runengine_action: QAction
    halt_runengine_action: QAction
    start_queue: QAction

    # Actions for showing window
    show_status_window_action: QtWidgets.QAction
    show_runs_window_action: QtWidgets.QAction
    show_energy_window_action: QtWidgets.QAction
    show_bss_window_action: QtWidgets.QAction
    launch_queuemonitor_action: QtWidgets.QAction

    # Keep track of motors
    motor_actions: Sequence = []
    motor_window_slots: Sequence = []

    # Keep track of cameras
    camera_actions: Mapping = {}
    camera_window_slots: Sequence

    # Keep track of area detectors
    area_detector_actions: Mapping = {}
    area_detector_window_slots: Sequence

    # Keep track of XRF detectors
    xrf_detector_actions: Mapping = {}
    xrf_detector_window_slots: Sequence

    # Signals for running plans on the queueserver
    queue_item_added = Signal(object)

    # Signals responding to queueserver changes
    queue_length_changed = Signal(int)

    def __init__(self, display="status", use_main_window=False, *args, **kwargs):
        # Instantiate the parent class
        # (*ui_file* and *use_main_window* let us render the window here instead)
        self.default_display = display
        super().__init__(ui_file=None, use_main_window=use_main_window, *args, **kwargs)
        self.windows = OrderedDict()

    def __del__(self):
        if hasattr(self, "_queue_thread"):
            self._queue_thread.quit()

    def _setup_window_action(self, action_name: str, text: str, slot: QtCore.Slot):
        action = QtWidgets.QAction(self)
        action.setObjectName(action_name)
        action.setText(text)
        action.triggered.connect(slot)
        setattr(self, action_name, action)

    def load_instrument(self):
        # Define devices on the beamline
        haven.load_instrument()
        # Make actions for launching other windows
        self.setup_window_actions()
        # Actions for controlling the bluesky run engine
        self.setup_runengine_actions()
        # Prepare the client for interacting with the queue server
        self.prepare_queue_client()
        # Launch the default display
        show_default_window = getattr(self, f"show_{self.default_display}_window")
        default_window = show_default_window()
        # Set up the window to show list of PV connections
        pydm.utilities.shortcuts.install_connection_inspector(parent=default_window)

    def setup_window_actions(self):
        """Create QActions for clicking on menu items, shortcuts, etc.

        These actions should be usable by multiple
        windows. Window-specific actions belong with the window.

        """
        self.prepare_motor_windows()
        self.prepare_camera_windows()
        self.prepare_area_detector_windows()
        self.prepare_xrf_detector_windows()
        # Action for showing the beamline status window
        self._setup_window_action(
            action_name="show_status_window_action",
            text="Beamline Status",
            slot=self.show_status_window,
        )
        # Action for showing the run browser window
        self._setup_window_action(
            action_name="show_run_browser_action",
            text="Browse Runs",
            slot=self.show_run_browser,
        )
        # Action for launch queue-monitor
        self._setup_window_action(
            action_name="launch_queuemonitor_action",
            text="Queue Monitor",
            slot=self.launch_queuemonitor,
        )
        # Action for showing the beamline status window
        self._setup_window_action(
            action_name="show_bss_window_action",
            text="Scheduling (&BSS)",
            slot=self.show_bss_window,
        )
        # Launch energy window
        self._setup_window_action(
            action_name="show_energy_window_action",
            text="Energy",
            slot=self.show_energy_window,
        )
        # Launch camera overview
        self._setup_window_action(
            action_name="show_cameras_window_action",
            text="All Cameras",
            slot=self.show_cameras_window,
        )

    def launch_queuemonitor(self):
        config = load_config()["queueserver"]
        zmq_info_addr = f"tcp://{config['info_host']}:{config['info_port']}"
        zmq_ctrl_addr = f"tcp://{config['control_host']}:{config['control_port']}"
        cmds = [
            "queue-monitor",
            "--zmq-control-addr",
            zmq_ctrl_addr,
            "--zmq-info-addr",
            zmq_info_addr,
        ]
        subprocess.Popen(cmds)

    def setup_runengine_actions(self):
        """Create QActions for controlling the bluesky runengine."""
        # Action for controlling the run engine
        actions = [
            ("pause_runengine_action", "Pause", "fa5s.stopwatch"),
            ("pause_runengine_now_action", "Pause Now", "fa5s.pause"),
            ("resume_runengine_action", "Resume", "fa5s.play"),
            ("stop_runengine_action", "Stop", "fa5s.stop"),
            ("abort_runengine_action", "Abort", "fa5s.eject"),
            ("halt_runengine_action", "Halt", "fa5s.ban"),
            ("start_queue_action", "Start", "fa5s.play"),
        ]
        for name, text, icon_name in actions:
            action = QtWidgets.QAction(self)
            icon = qta.icon(icon_name)
            action.setText(text)
            action.setIcon(icon)
            setattr(self, name, action)
            # action.triggered.connect(slot)

    def _prepare_device_windows(self, device_label: str, attr_name: str):
        """Generic routine to be called for individual classes of devices.

        Sets up window actions, windows and window slots for each
        instance of the this device class (specified by *device_label*).

        """
        try:
            devices = sorted(registry.findall(label=device_label), key=lambda x: x.name)
        except ComponentNotFound:
            log.warning(f"No {device_label} found, menu will be empty.")
            devices = []
        # Create menu actions for each device
        actions = {}
        setattr(self, f"{attr_name}_actions", actions)
        window_slots = []
        setattr(self, f"{attr_name}_window_slots", window_slots)
        setattr(self, f"{attr_name}_windows", {})
        for device in devices:
            action = QtWidgets.QAction(self)
            action.setObjectName(f"actionShow_{attr_name}_{device.name}")
            action.setText(device.name)
            actions[device.name] = action
            # Create a slot for opening the motor window
            slot = getattr(self, f"show_{attr_name}_window")
            slot = partial(slot, device=device)
            action.triggered.connect(slot)
            window_slots.append(slot)

    def prepare_area_detector_windows(self):
        """Prepare the support for opening area detector windows."""
        self._prepare_device_windows(
            device_label="area_detectors", attr_name="area_detector"
        )

    def prepare_xrf_detector_windows(self):
        """Prepare support to open X-ray fluorescence detector windows."""
        self._prepare_device_windows(
            device_label="xrf_detectors", attr_name="xrf_detector"
        )

    def prepare_camera_windows(self):
        self._prepare_device_windows(device_label="cameras", attr_name="camera")

    def prepare_motor_windows(self):
        """Prepare the support for opening motor windows."""
        # Get active motors
        try:
            motors = sorted(registry.findall(label="motors"), key=lambda x: x.name)
        except ComponentNotFound:
            log.warning(
                "No motors found, [Positioners] -> [Motors] menu will be empty."
            )
            motors = []
        # Create menu actions for each motor
        self.motor_actions = []
        self.motor_window_slots = []
        self.motor_windows = {}
        for motor in motors:
            action = QtWidgets.QAction(self)
            action.setObjectName(f"actionShow_Motor_{motor.name}")
            action.setText(motor.name)
            self.motor_actions.append(action)
            # Create a slot for opening the motor window
            slot = partial(self.show_motor_window, motor=motor)
            action.triggered.connect(slot)
            self.motor_window_slots.append(slot)

    def prepare_queue_client(self, api=None):
        if api is None:
            config = load_config()["queueserver"]
            ctrl_addr = f"tcp://{config['control_host']}:{config['control_port']}"
            info_addr = f"tcp://{config['info_host']}:{config['info_port']}"
            api = REManagerAPI(zmq_control_addr=ctrl_addr, zmq_info_addr=info_addr)
        client = QueueClient(api=api)
        thread = QueueClientThread(client=client)
        client.moveToThread(thread)
        # Connect actions to slots for controlling the queueserver
        self.pause_runengine_action.triggered.connect(
            partial(client.request_pause, defer=True)
        )
        self.pause_runengine_now_action.triggered.connect(
            partial(client.request_pause, defer=False)
        )
        self.start_queue_action.triggered.connect(client.start_queue)
        # Connect signals to slots for executing plans on queueserver
        self.queue_item_added.connect(client.add_queue_item)
        # Connect signals/slots for queueserver state changes
        client.length_changed.connect(self.queue_length_changed)
        # Start the thread
        thread.start()
        # Save references to the thread and runner
        self._queue_client = client
        # assert not hasattr(self, "_queue_thread")
        self._queue_thread = thread

    def add_queue_item(self, item):
        log.debug(f"Application received item to add to queue: {item}")
        self.queue_item_added.emit(item)

    def connect_menu_signals(self, window):
        """Connects application-level signals to the associated slots.

        These signals should generally be applicable to multiple
        windows. If the signal and/or slot is specific to a given
        window, then it should be in that widnow's class definition
        and setup code.

        """
        window.actionShow_Log_Viewer.triggered.connect(self.show_log_viewer_window)
        window.actionShow_Xafs_Scan.triggered.connect(self.show_xafs_scan_window)
        window.actionShow_Voltmeters.triggered.connect(self.show_voltmeters_window)
        window.actionShow_Sample_Viewer.triggered.connect(
            self.show_sample_viewer_window
        )

    def show_window(self, WindowClass, ui_file, name=None, macros={}):
        # Come up with the default key for saving in the windows dictionary
        if name is None:
            name = f"{WindowClass.__name__}_{ui_file.name}"
        # Check if the window has already been created
        if (w := self.windows.get(name)) is None:
            # Window is not yet created, so create one
            w = self.create_window(WindowClass, ui_dir / ui_file, macros=macros)
            self.windows[name] = w
            # Connect signals to remove the window when it closes
            w.destroyed.connect(partial(self.forget_window, name=name))
        else:
            # Window already exists so just bring it to the front
            w.show()
            w.activateWindow()
        return w

    def forget_window(self, obj, name):
        """Forget this window exists."""
        if hasattr(self, "windows"):
            del self.windows[name]

    def create_window(self, WindowClass, ui_file, macros={}):
        # Create and save this window
        main_window = WindowClass(
            hide_menu_bar=self.hide_menu_bar, hide_status_bar=self.hide_status_bar
        )
        # Make it look pretty
        apply_stylesheet(self.stylesheet_path, widget=main_window)
        main_window.update_tools_menu()
        # Load the UI file for this window
        display = main_window.open(str(ui_file.resolve()), macros=macros)
        self.connect_menu_signals(window=main_window)
        # Show the display
        if self.fullscreen:
            main_window.enter_fullscreen()
        else:
            main_window.show()
        return main_window

    def show_motor_window(self, *args, motor: HavenMotor):
        """Instantiate a new main window for this application."""
        motor_name = motor.name.replace(" ", "_")
        self.show_window(
            FireflyMainWindow,
            ui_dir / "motor.py",
            name=f"FireflyMainWindow_motor_{motor_name}",
            macros={"MOTOR": motor.name},
        )

    def show_camera_window(self, *args, device):
        """Instantiate a new main window for this application."""
        device_name = device.name.replace(" ", "_")
        self.show_window(
            FireflyMainWindow,
            ui_dir / "area_detector_viewer.py",
            name=f"FireflyMainWindow_camera_{device_name}",
            macros={"AD": device.name},
        )

    def show_area_detector_window(self, *args, device):
        """Instantiate a new main window for this application."""
        device_name = device.name.replace(" ", "_")
        self.show_window(
            FireflyMainWindow,
            ui_dir / "area_detector_viewer.py",
            name=f"FireflyMainWindow_area_detector_{device_name}",
            macros={"AD": device.name},
        )

    def show_xrf_detector_window(self, *args, device):
        """Instantiate a new main window for this application."""
        device_name = device.name.replace(" ", "_")
        self.show_window(
            FireflyMainWindow,
            ui_dir / "xrf_detector.py",
            name=f"FireflyMainWindow_xrf_detector_{device_name}",
            macros={"DEV": device.name},
        )

    def show_status_window(self, stylesheet_path=None):
        """Instantiate a new main window for this application."""
        self.show_window(
            FireflyMainWindow, ui_dir / "status.py", name="beamline_status"
        )

    @QtCore.Slot()
    def show_run_browser(self):
        return self.show_window(
            PlanMainWindow, ui_dir / "run_browser.py", name="run_browser"
        )

    @QtCore.Slot()
    def show_log_viewer_window(self):
        self.show_window(FireflyMainWindow, ui_dir / "log_viewer.ui", name="log_viewer")

    @QtCore.Slot()
    def show_xafs_scan_window(self):
        return self.show_window(
            PlanMainWindow, ui_dir / "xafs_scan.py", name="xafs_scan"
        )

    @QtCore.Slot()
    def show_voltmeters_window(self):
        return self.show_window(
            FireflyMainWindow, ui_dir / "voltmeters.py", name="voltmeters"
        )

    @QtCore.Slot()
    def show_sample_viewer_window(self):
        return self.show_window(
            FireflyMainWindow, ui_dir / "sample_viewer.ui", name="sample_viewer"
        )

    @QtCore.Slot()
    def show_cameras_window(self):
        return self.show_window(
            FireflyMainWindow, ui_dir / "cameras.py", name="cameras"
        )

    @QtCore.Slot()
    def show_energy_window(self):
        return self.show_window(PlanMainWindow, ui_dir / "energy.py", name="energy")

    @QtCore.Slot()
    def show_bss_window(self):
        return self.show_window(FireflyMainWindow, ui_dir / "bss.py", name="bss")



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import PVGroup, SubGroup, ioc_arg_parser, pvproperty, run

from haven.simulated_ioc import ResponsiveMotorFields


class MotorGroup(PVGroup):
    """
    An IOC with some motor records, similar to those found in a VME crate.

    E.g. "25idcVME:m1.VAL"

    """

    m1 = pvproperty(value=5000.0, doc="SLT V Upper", record=ResponsiveMotorFields)
    m2 = pvproperty(value=5000.0, doc="SLT V Lower", record=ResponsiveMotorFields)
    m3 = pvproperty(value=5000.0, doc="SLT H Inb", record=ResponsiveMotorFields)
    m4 = pvproperty(value=5000.0, doc="SLT H Otb", record=ResponsiveMotorFields)


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="225idVME:", desc="haven.tests.ioc_motor test IOC"
    )
    ioc = MotorGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import (
    PVGroup,
    SubGroup,
    ioc_arg_parser,
    pvproperty,
    run,
    PvpropertyDouble,
)
from ophyd.tests.fake_motor_ioc import FakeMotorIOC

from haven.simulated_ioc import ResponsiveMotorFields  # , IOC as IOC_


class SimpleGroup(PVGroup):
    """
    An IOC with three uncoupled read/writable PVs

    Scalar PVs
    ----------
    A (int)
    B (float)

    Vectors PVs
    -----------
    C (vector of int)
    """

    A = pvproperty(value=1, doc="An integer")
    B = pvproperty(value=2.0, doc="A float")
    C = pvproperty(value=[1, 2, 3], doc="An array of integers")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="simple:", desc="haven.tests.ioc_undulator test IOC"
    )
    ioc = SimpleGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



import logging
import warnings

from apstools.devices.aps_machine import ApsMachineParametersDevice
from apsbss.apsbss_ophyd import EpicsBssDevice

from haven import registry
from .._iconfig import load_config


log = logging.getLogger(__name__)


class ApsMachine(ApsMachineParametersDevice):
    _default_read_attrs = [
        "current",
        "lifetime",
    ]
    _default_configuration_attrs = [
        "aps_cycle",
        "machine_status",
        "operating_mode",
        "shutter_permit",
        "fill_number",
        "orbit_correction",
        "global_feedback",
        "global_feedback_h",
        "global_feedback_v",
        "operator_messages",
    ]


def load_aps(config=None):
    """Load devices related to the synchrotron as a whole."""
    devices = []
    if config is None:
        config = load_config()
    # Load storage ring device
    try:
        aps_ = ApsMachine(name="APS", labels={"synchrotrons"})
    except Exception as exc:
        msg = f"Could not instantiate APS machine: {repr(exc)}"
        log.warning(msg)
        warnings.warn(msg)
    else:
        registry.register(aps_)
        devices.append(aps_)
    # Load scheduling system device
    bss_ = EpicsBssDevice(prefix=f"{config['bss']['prefix']}:", name="bss")
    registry.register(bss_)
    devices.append(bss_)
    return devices



from enum import IntEnum
from collections import OrderedDict
from typing import Optional, Sequence
import warnings
import logging

from ophyd import (
    mca,
    Device,
    EpicsSignal,
    EpicsSignalRO,
    Component as Cpt,
    DynamicDeviceComponent as DDC,
    Kind,
)
from apstools.utils import cleanupText

from .scaler_triggered import ScalerTriggered
from .instrument_registry import registry
from .device import RegexComponent as RECpt
from .._iconfig import load_config
from .. import exceptions


__all__ = ["DxpDetectorBase", "load_fluorescence_detectors", "load_dxp_detector"]


log = logging.getLogger(__name__)


active_kind = Kind.normal | Kind.config


class ROI(mca.ROI):
    _default_read_attrs = [
        "count",
        "net_count",
    ]
    _default_configuration_attrs = [
        "label",
        "bkgnd_chans",
        "hi_chan",
        "lo_chan",
    ]
    # hints = {"fields": ["net_count"]}
    kind = active_kind
    _original_name = None
    _original_kinds = {}
    _dynamic_hint_fields = ["net_count"]
    # Signals
    # net_count = Cpt(EpicsSignalRO, "N", kind=Kind.hinted, lazy=True)
    # user_kind = Cpt(EpicsSignal, "_BS_KIND", lazy=True)
    is_hinted = RECpt(EpicsSignal, "BH", pattern=r"\.R", repl="_R", lazy=True)

    def stage(self):
        self._original_name = self.name
        # Append the ROI label to the signal name
        label = cleanupText(str(self.label.get()))
        if label != "":
            self.name = f"{self.name}_{label}"
        # Set the kind based on the user-settable ".R0_BS_HINTED" PV
        if self.is_hinted.get():
            new_kind = Kind.hinted
        else:
            new_kind = Kind.normal
        self._original_kinds = {
            fld: getattr(self, fld).kind for fld in self._dynamic_hint_fields
        }
        for fld in self._dynamic_hint_fields:
            getattr(self, fld).kind = new_kind
        super().stage()

    def unstage(self):
        # Restore the original (pre-staged) name
        self.name = self._original_name
        # Restore original signal kinds
        for fld, kind in self._original_kinds.items():
            getattr(self, fld).kind = kind
        super().unstage()


def add_rois(range_: Sequence[int] = range(32), kind=Kind.normal, **kwargs):
    """Add one or more ROIs to an MCA instance

    Parameters
    ----------
    range_ : sequence of ints
        Must be be in the set [0,31]

    By default, an EpicsMCA is initialized with all 32 rois. These
    provide the following Components as EpicsSignals (N=[0,31]):
    EpicsMCA.rois.roiN.(label,count,net_count,preset_cnt, is_preset,
    bkgnd_chans, hi_chan, lo_chan)

    """
    defn = OrderedDict()
    kwargs["kind"] = kind
    for roi in range_:
        if not (0 <= roi < 32):
            raise ValueError("roi must be in the set [0,31]")
        attr = "roi{}".format(roi)
        defn[attr] = (
            ROI,
            ".R{}".format(roi),
            kwargs,
        )
    return defn


class MCARecord(mca.EpicsMCARecord):
    rois = DDC(add_rois(), kind=active_kind)
    _default_read_attrs = [
        "rois",
        "spectrum",
        "preset_real_time",
        "preset_live_time",
        "elapsed_real_time",
        "elapsed_live_time",
        "background",
    ]
    _default_configuration_attrs = ["rois", "mode"]
    kind = active_kind


def add_mcas(range_, kind=active_kind, **kwargs):
    """Add one or more MCARecords to a device

    Parameters
    ----------
    range_
      Indices for which to create MCA records.

    """
    defn = OrderedDict()
    kwargs["kind"] = kind
    for idx in range_:
        attr = f"mca{idx}"
        defn[attr] = (
            MCARecord,
            f"mca{idx}",
            kwargs,
        )
    return defn


class DxpDetectorBase(mca.EpicsDXPMultiElementSystem):
    """A fluorescence detector based on XIA-DXP XMAP electronics.

    Creates MCA components based on the number of elements.

    """

    # By default, a 1-element detector, subclass for more elements
    mcas = DDC(
        add_mcas(range_=range(1, 3)),
        default_read_attrs=["mca1"],
        default_configuration_attrs=["mca1"],
    )
    _default_read_attrs = [
        "preset_live_time",
        "preset_real_time",
        "dead_time",
        "elapsed_live",
        "elapsed_real",
        "mcas",
    ]
    _default_configuration_attrs = [
        "max_scas",
        "num_scas",
        "poll_time",
        "prescale",
        "preset_mode",
        "preset_events",
        "preset_triggers",
        "snl_connected",
        "mcas",
    ]
    _omitted_attrs = [  # Just save them here for easy reference
        "channel_advance",
        "client_wait",
        "dwell",
        "save_system",
        "save_system_file",
        "set_client_wait",
        "erase_all",
        "erase_start",
        "start_all",
        "stop_all",
        "set_acquire_busy",
        "acquire_busy",
        "status_all",
        "status_all_once",
        "acquiring",
        "read_baseline_histograms",
        "read_all",
        "read_all_once",
        "copy_adcp_ercent_rule",
        "copy_baseline_cut_enable",
        "copy_baseline_cut_percent",
        "copy_baseline_filter_length",
        "copy_baseline_threshold",
        "copy_decay_time",
        "copy_detector_polarity",
        "copy_energy_threshold",
        "copy_gap_time",
        "copy_max_energy",
        "copy_max_width",
        "copy_peaking_time",
        "copy_preamp_gain",
        "copy_roic_hannel",
        "copy_roie_nergy",
        "copy_roi_sca",
        "copy_reset_delay",
        "copy_trigger_gap_time",
        "copy_trigger_peaking_time",
        "copy_trigger_threshold",
        "do_read_all",
        "do_read_baseline_histograms",
        "do_read_traces",
        "do_status_all",
        "read_low_level_params",
        "read_traces",
        "trace_modes",
        "trace_times",
        "idead_time",
    ]

    @property
    def num_rois(self):
        n_rois = float("inf")
        for mca in self.mca_records():
            n_rois = min(n_rois, len(mca.rois.component_names))
        return n_rois

    @property
    def num_elements(self):
        return len(self.mca_records())

    def mca_records(self, mca_indices: Optional[Sequence[int]] = None):
        mcas = [
            getattr(self.mcas, m)
            for m in self.mcas.component_names
            if m.startswith("mca")
        ]
        # Filter by element index
        if mca_indices is not None:
            mcas = [
                m for m in mcas if int(m.dotted_name.split(".")[-1][3:]) in mca_indices
            ]
        return mcas

    def rois(self, roi_indices: Optional[Sequence[int]] = None):
        # Get the list of ROIs to activate
        all_rois = []
        for mca in self.mca_records():
            rois = mca.rois.component_names
            rois = [getattr(mca.rois, r) for r in rois]
            # Get sub-list if requested
            if roi_indices is not None:
                rois = [rois[i] for i in roi_indices]
            all_rois.extend(rois)
        return all_rois

    def enable_rois(self, rois: Optional[Sequence[int]] = None):
        """Include some, or all, ROIs in the list of detectors to
        read.

        rois
          A list of indices for which ROIs to enable. Default is to
          operate on all ROIs.

        """
        for roi in self.rois(roi_indices=rois):
            roi.kind = active_kind

    def disable_rois(self, rois: Optional[Sequence[int]] = None):
        """Remove some, or all, ROIs from the list of detectors to
        read.

        rois
          A list of indices for which ROIs to enable. Default is to
          operate on all ROIs.

        """
        for roi in self.rois(roi_indices=rois):
            roi.kind = Kind.omitted

    def enable_elements(self, elements: Optional[Sequence[int]] = None):
        """Include some, or all, elements in the list of detectors to
        read.

        elements
          A list of indices for which elements to enable. Default is
          to operate on all elements.

        """
        for elem in self.mca_records(mca_indices=elements):
            elem.kind = active_kind

    def disable_elements(self, elements: Optional[Sequence[int]] = None):
        """Include some, or all, elements in the list of detectors to
        read.

        elements
          A list of indices for which elements to enable. Default is
          to operate on all elements.

        """
        for elem in self.mca_records(mca_indices=elements):
            elem.kind = Kind.omitted


class XspressDetector(ScalerTriggered, Device):
    """A fluorescence detector plugged into an Xspress3 readout."""

    num_frames = Cpt(EpicsSignal, "NumImages", kind=Kind.config)
    trigger_mode = Cpt(EpicsSignal, "TriggerMode", kind=Kind.config)
    acquire = Cpt(EpicsSignal, "Acquire", kind=Kind.omitted)
    erase = Cpt(EpicsSignal, "Erase", kind=Kind.omitted)

    class erase_states(IntEnum):
        DONE = 0
        ERASE = 1

    class acquire_states(IntEnum):
        DONE = 0
        ACQUIRE = 1

    class mode(IntEnum):
        SOFTWARE = 0
        INTERNAL = 1
        IDC = 2
        TTL_VETO_ONLY = 3
        TTL_BOTH = 4
        LVDS_VETO_ONLY = 5
        LVDS_BOTH = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs[self.erase] = self.erase_states.ERASE
        self.stage_sigs[self.trigger_mode] = self.mode.TTL_VETO_ONLY
        self.stage_sigs[self.acquire] = self.acquire_states.ACQUIRE

    @property
    def stage_num_frames(self):
        """How many frames to prepare for when staging this detector."""
        return self.stage_sigs.get(self.num_frames, 1)

    @stage_num_frames.setter
    def stage_num_frames(self, val):
        self.stage_sigs[self.num_frames] = val


def load_dxp_detector(device_name, prefix, num_elements):
    # Build the mca components
    # (Epics uses 1-index instead of 0-index)
    mca_range = range(1, num_elements + 1)
    attrs = {
        "mcas": DDC(
            add_mcas(range_=mca_range),
            kind=active_kind,
            default_read_attrs=[f"mca{i}" for i in mca_range],
            default_configuration_attrs=[f"mca{i}" for i in mca_range],
        )
    }
    # Create a dynamic subclass with the MCAs
    class_name = device_name.title().replace("_", "")
    parent_classes = (DxpDetectorBase,)
    Cls = type(class_name, parent_classes, attrs)
    det = Cls(prefix=f"{prefix}:", name=device_name, labels={"xrf_detectors"})
    registry.register(det)


def load_fluorescence_detectors(config=None):
    # Get the detector definitions from config files
    if config is None:
        config = load_config()
    for name, cfg in config.get("fluorescence_detector", {}).items():
        if "prefix" not in cfg.keys():
            continue
        # Build the detector device
        if cfg["electronics"] == "dxp":
            load_dxp_detector(
                device_name=name, prefix=cfg["prefix"], num_elements=cfg["num_elements"]
            )
        else:
            msg = f"Electronics '{cfg['electronics']}' for {name} not supported."
            raise exceptions.UnknownDeviceConfiguration(msg)



import databroker
from tiled.client import from_uri
from tiled.client.cache import Cache

from ._iconfig import load_config


def load_catalog(name: str = "bluesky"):
    """Load a databroker catalog for retrieving data.

    To retrieve individual scans, consider the ``load_result`` and
    ``load_data`` methods.

    Parameters
    ==========
    name
      The name of the catalog as defined in the Intake file
      (e.g. ~/.local/share/intake/catalogs.yml)

    Returns
    =======
    catalog
      The databroker catalog.
    """
    return databroker.catalog[name]


def load_result(uid: str, catalog_name: str = "bluesky", stream: str = "primary"):
    """Load a past experiment from the database.

    The result contains metadata and scan parameters. The data
    themselves are accessible from the result's *read()* method.

    Parameters
    ==========
    uid
      The universal identifier for this scan, as return by a bluesky
      RunEngine.
    catalog_name
      The name of the catalog as defined in the Intake file
      (e.g. ~/.local/share/intake/catalogs.yml)
    stream
      The data stream defined by the bluesky RunEngine.

    Returns
    =======
    result
      The experiment result, with data available via the *read()*
      method.

    """
    cat = load_catalog(name=catalog_name)
    result = cat[uid][stream]
    return result


def load_data(uid, catalog_name="bluesky", stream="primary"):
    """Load a past experiment's data from the database.

    The result is an xarray with the data collected.

    Parameters
    ==========
    uid
      The universal identifier for this scan, as return by a bluesky
      RunEngine.
    catalog_name
      The name of the catalog as defined in the Intake file
      (e.g. ~/.local/share/intake/catalogs.yml)
    stream
      The data stream defined by the bluesky RunEngine.

    Returns
    =======
    data
      The experimental data, as an xarray.

    """

    res = load_result(uid=uid, catalog_name=catalog_name, stream=stream)
    data = res.read()
    return data


def tiled_client(entry_node=None, uri=None):
    config = load_config()
    if uri is None:
        uri = config["database"]["tiled"]["uri"]
    client_ = from_uri(uri, cache=Cache.in_memory(2e9))
    if entry_node is None:
        entry_node = config["database"]["tiled"]["entry_node"]
    client_ = client_[entry_node]
    return client_



from haven.instrument.shutter import load_shutters
from haven import registry


def test_shutter():
    load_shutters()
    shutters = list(registry.findall(label="shutters"))
    assert len(shutters) == 2
    shutterA = registry.find(name="front_end_shutter")
    assert shutterA.name == "front_end_shutter"
    assert shutterA.open_signal.pvname == "PSS:99ID:FES_OPEN_EPICS.VAL"
    assert shutterA.close_signal.pvname == "PSS:99ID:FES_CLOSE_EPICS.VAL"
    assert shutterA.pss_state.pvname == "PSS:99ID:A_BEAM_PRESENT"
    shutterC = registry.find(name="hutch_shutter")
    assert shutterC.name == "hutch_shutter"
    assert shutterC.open_signal.pvname == "PSS:99ID:SCS_OPEN_EPICS.VAL"
    assert shutterC.close_signal.pvname == "PSS:99ID:SCS_CLOSE_EPICS.VAL"
    assert shutterC.pss_state.pvname == "PSS:99ID:C_BEAM_PRESENT"



from unittest.mock import MagicMock
from ophyd import Kind
from bluesky import plans as bp

import pytest

from epics import caget


def test_load_dxp(sim_registry, mocker):
    mocker.patch("ophyd.signal.EpicsSignalBase._ensure_connected")
    from haven.instrument.fluorescence_detector import load_fluorescence_detectors

    load_fluorescence_detectors(config=None)
    # See if the device was loaded
    vortex = sim_registry.find(name="vortex_me4")
    # Check that the MCA's are available
    assert hasattr(vortex.mcas, "mca1")
    assert hasattr(vortex.mcas, "mca2")
    assert hasattr(vortex.mcas, "mca3")
    assert hasattr(vortex.mcas, "mca4")
    # Check that MCA's have ROI's available
    assert hasattr(vortex.mcas.mca1, "rois")
    assert hasattr(vortex.mcas.mca1.rois, "roi1")
    # Check that bluesky hints were added
    assert hasattr(vortex.mcas.mca1.rois.roi1, "is_hinted")
    assert vortex.mcas.mca1.rois.roi1.is_hinted.pvname == "vortex_me4:mca1_R1BH"


@pytest.fixture()
def vortex(sim_registry, mocker):
    mocker.patch("ophyd.signal.EpicsSignalBase._ensure_connected")
    from haven.instrument.fluorescence_detector import load_fluorescence_detectors

    load_fluorescence_detectors(config=None)
    # See if the device was loaded
    vortex = sim_registry.find(name="vortex_me4")
    yield vortex


@pytest.mark.xfail
def test_enable_rois(vortex):
    assert hasattr(vortex.mcas.mca1, "rois")
    # Test that all channels are disabled by default
    assert "vortex_me4_mca1_rois_roi1" not in vortex.read_attrs
    assert "preset_live_time" in vortex.read_attrs
    assert "preset_mode" in vortex.configuration_attrs
    # Make sure config attrs are in the right place
    config_signals = [
        "mcas.mca3.rois.roi0.label",
        "mcas.mca3.rois.roi0.bkgnd_chans",
        "mcas.mca3.rois.roi0.hi_chan",
        "mcas.mca3.rois.roi0.lo_chan",
    ]
    normal_signals = ["mcas.mca3.rois.roi0.count", "mcas.mca3.rois.roi0.net_count"]
    omitted_signals = [
        "mcas.mca3.rois.roi0.preset_count",
        "mcas.mca3.rois.roi0.is_preset",
    ]
    for attr in config_signals:
        assert attr not in vortex.configuration_attrs
        assert attr not in vortex.read_attrs
    for attr in normal_signals:
        assert attr not in vortex.read_attrs
        assert attr not in vortex.configuration_attrs
    for attr in omitted_signals:
        assert attr not in vortex.read_attrs
        assert attr not in vortex.configuration_attrs
    # Enable all ROIs
    vortex.enable_rois()
    for attr in config_signals:
        assert attr in vortex.configuration_attrs
        assert attr not in vortex.read_attrs
    for attr in normal_signals:
        assert attr in vortex.read_attrs
        assert attr not in vortex.configuration_attrs
    for attr in omitted_signals:
        assert attr not in vortex.read_attrs
        assert attr not in vortex.configuration_attrs
    # Disable all ROIs
    vortex.disable_rois()
    for attr in config_signals:
        assert attr not in vortex.configuration_attrs
        assert attr not in vortex.read_attrs
    for attr in normal_signals:
        assert attr not in vortex.read_attrs
        assert attr not in vortex.configuration_attrs
    for attr in omitted_signals:
        assert attr not in vortex.read_attrs
        assert attr not in vortex.configuration_attrs


def test_enable_some_rois(vortex):
    vortex.disable_rois()
    assert "mcas.mca1.rois.roi0" not in vortex.read_attrs
    assert "mcas.mca1.rois.roi1" not in vortex.read_attrs
    # Enable just ROI0
    vortex.enable_rois(rois=[0])
    assert "mcas.mca1.rois.roi0" in vortex.read_attrs
    assert "mcas.mca1.rois.roi1" not in vortex.read_attrs
    # Disable just ROI0
    vortex.enable_rois()
    vortex.disable_rois(rois=[0])
    assert "mcas.mca1.rois.roi0" not in vortex.read_attrs
    assert "mcas.mca1.rois.roi1" in vortex.read_attrs


def test_enable_all_elements(vortex):
    vortex.enable_rois()
    vortex.disable_elements()
    assert "mcas.mca1.rois.roi0" not in vortex.read_attrs
    assert "mcas.mca1.rois.roi1" not in vortex.read_attrs
    # Enable just ROI0
    vortex.enable_elements(elements=[1])
    assert "mcas.mca1.rois.roi0" in vortex.read_attrs
    assert "mcas.mca2.rois.roi0" not in vortex.read_attrs
    # Disable just ROI0
    vortex.enable_elements()
    vortex.disable_elements(elements=[0])
    assert "mcas.mca0.rois.roi0" not in vortex.read_attrs
    assert "mcas.mca1.rois.roi0" in vortex.read_attrs


@pytest.mark.xfail
def test_with_plan(vortex):
    assert False, "Write test"


def test_stage_signal_names(sim_vortex):
    """Check that we can set the name of the detector ROIs dynamically."""
    dev = sim_vortex.mcas.mca1.rois.roi1
    dev.label.put("Ni-Ka")
    # Ensure the name isn't changed yet
    assert "Ni-Ka" not in dev.name
    assert "Ni_Ka" not in dev.name
    orig_name = dev.name
    dev.stage()
    try:
        result = dev.read()
    except Exception:
        raise
    else:
        assert "Ni-Ka" not in dev.name  # Make sure it gets sanitized
        assert "Ni_Ka" in dev.name
    finally:
        dev.unstage()
    # Name gets reset when unstaged
    assert dev.name == orig_name
    # Check acquired data uses dynamic names
    for res in result.keys():
        assert "Ni_Ka" in res


def test_stage_signal_hinted(sim_vortex):
    dev = sim_vortex.mcas.mca1.rois.roi1
    # Check that ROI is not hinted by default
    assert dev.name not in sim_vortex.hints
    # Enable the ROI by setting it's kind PV to "hinted"
    dev.is_hinted.put(True)
    # Ensure signals are not hinted before being staged
    assert dev.net_count.name not in sim_vortex.hints["fields"]
    try:
        dev.stage()
    except Exception:
        raise
    else:
        assert dev.net_count.name in sim_vortex.hints["fields"]
        assert (
            sim_vortex.mcas.mca1.rois.roi0.net_count.name
            not in sim_vortex.hints["fields"]
        )
    finally:
        dev.unstage()
    # Did it restore kinds properly when unstaging
    assert dev.net_count.name not in sim_vortex.hints["fields"]
    assert (
        sim_vortex.mcas.mca1.rois.roi0.net_count.name not in sim_vortex.hints["fields"]
    )



from ophyd import DetectorBase

from haven import registry, load_config
from haven.instrument.camera import AravisDetector, load_cameras


def test_load_cameras(ioc_camera):
    load_cameras(config=load_config())
    # Check that cameras were registered
    cameras = list(registry.findall(label="cameras"))
    assert len(cameras) == 1
    assert isinstance(cameras[0], DetectorBase)


def test_camera_device(ioc_camera):
    camera = AravisDetector(ioc_camera.prefix, name="camera")
    assert isinstance(camera, DetectorBase)
    assert hasattr(camera, "cam")


def test_camera_in_registry(sim_registry, ioc_camera):
    camera = AravisDetector(ioc_camera.prefix, name="camera")
    sim_registry.register(camera)
    # Check that all sub-components are accessible
    camera = sim_registry.find(camera.name)
    cam = sim_registry.find(f"{camera.name}_cam")
    gain = sim_registry.find(f"{camera.name}_cam.gain")



from enum import IntEnum

from ophyd import (
    Device,
    Component as Cpt,
    FormattedComponent as FCpt,
    EpicsMotor,
    EpicsSignal,
    EpicsSignalRO,
)

from .instrument_registry import registry
from .._iconfig import load_config


class IDTracking(IntEnum):
    OFF = 0
    ON = 1


@registry.register
class Monochromator(Device):
    # ID tracking PVs
    id_tracking = Cpt(EpicsSignal, ":ID_tracking", kind="config")
    id_offset = Cpt(EpicsSignal, ":ID_offset", kind="config")
    d_spacing = Cpt(EpicsSignal, ":dspacing", kind="config")
    # Virtual positioners
    mode = Cpt(EpicsSignal, ":mode", labels={"motors", "baseline"}, kind="config")
    energy = Cpt(EpicsMotor, ":Energy", labels={"motors"}, kind="hinted")
    energy_constant1 = Cpt(
        EpicsSignal, ":EnergyC1.VAL", labels={"baseline"}, kind="config"
    )
    energy_constant2 = Cpt(
        EpicsSignal, ":EnergyC2.VAL", labels={"baseline"}, kind="config"
    )
    energy_constant3 = Cpt(
        EpicsSignal, ":EnergyC3.VAL", labels={"baseline"}, kind="config"
    )
    offset = Cpt(EpicsMotor, ":Offset", labels={"motors", "baseline"}, kind="config")
    # ACS Motors
    horiz = Cpt(EpicsMotor, ":ACS:m1", labels={"motors", "baseline"}, kind="config")
    vert = Cpt(EpicsMotor, ":ACS:m2", labels={"motors", "baseline"}, kind="config")
    bragg = Cpt(EpicsMotor, ":ACS:m3", labels={"motors"})
    gap = Cpt(EpicsMotor, ":ACS:m4", labels={"motors"})
    roll2 = Cpt(EpicsMotor, ":ACS:m5", labels={"motors", "baseline"}, kind="config")
    pitch2 = Cpt(EpicsMotor, ":ACS:m6", labels={"motors", "baseline"}, kind="config")
    roll_int = Cpt(EpicsMotor, ":ACS:m7", labels={"motors", "baseline"}, kind="config")
    pi_int = Cpt(EpicsMotor, ":ACS:m8", labels={"motors", "baseline"}, kind="config")
    # Physical constants
    d_spacing = Cpt(EpicsSignalRO, ":dspacing", labels={"baseline"}, kind="config")


def load_monochromator(config=None):
    # Load PV's from config
    if config is None:
        config = load_config()
    monochromator = Monochromator(config["monochromator"]["ioc"], name="monochromator")
    return monochromator



from ophyd.device import do_not_wait_for_lazy_connection

from haven.instrument.area_detector import load_area_detectors


def test_load_area_detectors(sim_registry, ioc_area_detector):
    load_area_detectors()
    # Check that some area detectors were loaded
    dets = sim_registry.findall(label="area_detectors")



#!/usr/bin/env python3
from caproto import ChannelType
from caproto.server import (
    PVGroup,
    SubGroup,
    ioc_arg_parser,
    pvproperty,
    run,
    PvpropertyDouble,
)
from ophyd.tests.fake_motor_ioc import FakeMotorIOC

from haven.simulated_ioc import ResponsiveMotorFields  # , IOC as IOC_


class UndulatorGroup(PVGroup):

    """
    An IOC that looks like an undulator.

    E.g. "25IDds:Energy"

    """

    ScanEnergy = pvproperty(value=0, doc="ID Energy Scan Input", dtype=PvpropertyDouble)
    Energy = pvproperty(
        value=0, doc="", record=ResponsiveMotorFields, dtype=PvpropertyDouble
    )
    Busy = pvproperty(value=0, doc="", name="Busy.VAL")
    Stop = pvproperty(value=0, doc="", name="Stop.VAL")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255ID:", desc="haven.tests.ioc_undulator test IOC"
    )
    ioc = UndulatorGroup(**ioc_options)
    run(ioc.pvdb, **run_options)



"""The scaler has a global trigger, but individual channels.

These tests check the system for monitoring the trigger system in
bluesky so that it only fires the trigger once for all of the scaler
channels. Additionally, the scaler can send that trigger to other
devices, like the Xspress3 fluorescence detector readout electronics.

"""

import pytest
from ophyd import Device


from haven.instrument.scaler_triggered import ScalerTriggered


@pytest.mark.skip("scaler_triggering is not needed right now")
def test_trigger_fires():
    scaler_prefix = "myioc:myscaler"
    # Prepare the scaler triggered device
    MyDevice = type("MyDevice", (ScalerTriggered, Device), {})
    device = MyDevice(name="device", scaler_prefix=scaler_prefix)
    # Make sure it's in a sensible starting state
    assert hasattr(device, "_statuses")
    assert len(device._statuses) == 0
    # Trigger the device
    device.trigger()
    # Check that a status was added
    assert len(device._statuses) == 1
    # Check that triggering again is indempotent
    first_status = device._statuses[scaler_prefix]
    old_done = first_status.__class__.done
    first_status.__class__.done = False
    try:
        device.trigger()
    except Exception:
        pass
    finally:
        first_status.__class__.done = old_done
    second_status = device._statuses[scaler_prefix]
    assert first_status is second_status



import numpy as np
from ophyd.pseudopos import pseudo_position_argument, real_position_argument
from ophyd import PseudoPositioner, PseudoSingle, EpicsMotor
from ophyd import Component as Cpt, FormattedComponent as FCpt, Device

from .._iconfig import load_config
from .instrument_registry import registry


um_per_mm = 1000


class RowlandPositioner(PseudoPositioner):
    """A pseudo positioner describing a rowland circle.

    Real Axes
    =========
    x
    y
    z
    z1

    Pseudo Axes
    ===========
    D
      In mm
    theta
      In degrees
    alpha
      In degrees
    """

    def __init__(
        self,
        x_motor_pv: str,
        y_motor_pv: str,
        z_motor_pv: str,
        z1_motor_pv: str,
        *args,
        **kwargs
    ):
        self.x_motor_pv = x_motor_pv
        self.y_motor_pv = y_motor_pv
        self.z_motor_pv = z_motor_pv
        self.z1_motor_pv = z1_motor_pv
        super().__init__(*args, **kwargs)

    # Pseudo axes
    D: PseudoSingle = Cpt(PseudoSingle, name="D", limits=(0, 1000))
    theta: PseudoSingle = Cpt(PseudoSingle, name="theta", limits=(0, 180))
    alpha: PseudoSingle = Cpt(PseudoSingle, name="alpha", limits=(0, 180))

    # Real axes
    x: EpicsMotor = FCpt(EpicsMotor, "{x_motor_pv}", name="x")
    y: EpicsMotor = FCpt(EpicsMotor, "{y_motor_pv}", name="y")
    z: EpicsMotor = FCpt(EpicsMotor, "{z_motor_pv}", name="z")
    z1: EpicsMotor = FCpt(EpicsMotor, "{z1_motor_pv}", name="z1")

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Run a forward (pseudo -> real) calculation"""
        # Convert distance to microns and degrees to radians
        D = pseudo_pos.D * um_per_mm
        theta = pseudo_pos.theta / 180.0 * np.pi
        alpha = pseudo_pos.alpha / 180.0 * np.pi
        # Convert virtual positions to real positions
        x = D * (np.sin(theta + alpha)) ** 2
        y = D * ((np.sin(theta + alpha)) ** 2 - (np.sin(theta - alpha)) ** 2)
        z1 = D * np.sin(theta - alpha) * np.cos(theta + alpha)
        z2 = D * np.sin(theta - alpha) * np.cos(theta - alpha)
        z = z1 + z2
        return self.RealPosition(
            x=x,
            y=y,
            z=z,
            z1=z1,
        )

    @real_position_argument
    def inverse(self, real_pos):
        """Run an inverse (real -> pseudo) calculation"""
        return self.PseudoPosition(D=0, theta=0, alpha=0)
        # Expand the variables
        x = real_pos.x
        y = real_pos.y
        z = real_pos.z
        z1 = real_pos.z1
        # Invert the calculation, first for 'd'
        a = y
        b = -x
        c = z1**2
        d = -(z1**2 * y)  # Maybe needs parens?
        p = (3 * a * c - b**2) / (3 * a**2)
        q = (2 * b**3 - 9 * a * b * c + 27 * a**2 * d) / (27 * a**3)
        D = (-(q / 2) + (q**2 / 4 + p**3 / 27) ** 0.5) ** (1 / 3) + (
            -(q / 2) - (q**2 / 4 + p**3 / 27) ** 0.5
        ) ** (1 / 3)
        # D = x / ((1 - (z1 ** 2 / (D ** 2 * y + z1 ** 2))) ** 2)
        # cos(theta + alpha) = (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5
        print(a, b, c, d)
        print(p, q, D)
        return self.PseudoPosition(D=D, theta=0, alpha=0)


# Rewrite the following four equations so that they calculate D, theta and alpha based on inputting x, y, z1 and z.

#         x = D * (sin(theta + alpha)) ** 2
#         y = D * ((sin(theta + alpha)) ** 2 - (sin(theta - alpha)) ** 2)
#         z1 = D * sin(theta - alpha) * cos(theta + alpha)
#         z = z1 + D * sin(theta - alpha) * cos(theta - alpha)

# To calculate D, theta, and alpha based on x, y, z1, and z, we can rearrange the given equations as follows:

# D = x / (sin(theta + alpha)) ** 2

# sin(theta - alpha) = z1 / (D * cos(theta + alpha))

# cos(theta - alpha) = (z - z1) / (D * sin(theta - alpha))

# Substitute the above two equations into the expression for y:

# y = D * ((sin(theta + alpha)) ** 2 - (sin(theta - alpha)) ** 2)

# y = D * ((sin(theta + alpha)) ** 2 - (z1 / (D * cos(theta + alpha))) ** 2)

# y = D * ((sin(theta + alpha)) ** 2 - (z1 ** 2 / D ** 2) * (cos(theta + alpha)) ** (-2))

# Solve for cos(theta + alpha) in the above equation:

# cos(theta + alpha) = (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5

# Substitute the value of cos(theta + alpha) into the equation for sin(theta - alpha):

# sin(theta - alpha) = z1 / (D * (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5)

# Simplify the above equation:

# sin(theta - alpha) = (z1 ** 2 / (D ** 2 * y + z1 ** 2)) ** 0.5

# Finally, substitute the values of cos(theta + alpha) and sin(theta - alpha) into the equation for D:

# D = x / ((sin(theta + alpha)) ** 2)

# D = x / ((1 - (z1 ** 2 / (D ** 2 * y + z1 ** 2))) ** 2)

# This equation can be solved numerically to obtain the value of D. Once D is known, we can use the equations for cos(theta + alpha) and sin(theta - alpha) to calculate theta and alpha.


@registry.register
class LERIXSpectrometer(Device):
    rowland = Cpt(
        RowlandPositioner,
        x_motor_pv="vme_crate_ioc:m1",
        y_motor_pv="vme_crate_ioc:m2",
        z_motor_pv="vme_crate_ioc:m3",
        z1_motor_pv="vme_crate_ioc:m4",
        name="rowland",
    )


def load_lerix_spectrometers(config=None):
    if config is None:
        config = load_config()
    # Create spectrometers
    for name, cfg in config.get("lerix", {}).items():
        rowland = cfg["rowland"]
        device = RowlandPositioner(
            name=name,
            x_motor_pv=rowland["x_motor_pv"],
            y_motor_pv=rowland["y_motor_pv"],
            z_motor_pv=rowland["z_motor_pv"],
            z1_motor_pv=rowland["z1_motor_pv"],
            labels={"lerix_spectrometers"},
        )
        registry.register(device)



import epics

from haven.instrument.instrument_registry import registry
from haven.instrument import motor


def test_load_vme_motors(ioc_motor):
    registry.clear()
    # Set the IOC motor descriptions to known values
    epics.caput("255idVME:m1.DESC", "SLT V Upper")
    epics.caput("255idVME:m2.DESC", "SLT V Lower")
    epics.caput("255idVME:m3.DESC", "SLT H Inbound")
    assert epics.caget("255idVME:m1.DESC", use_monitor=False) == "SLT V Upper"
    assert epics.caget("255idVME:m2.DESC", use_monitor=False) == "SLT V Lower"
    assert epics.caget("255idVME:m3.DESC", use_monitor=False) == "SLT H Inbound"
    # Load the Ophyd motor definitions
    motor.load_all_motors()
    # Were the motors imported correctly
    motors = list(registry.findall(label="motors"))
    assert len(motors) == 3
    assert type(motors[0]) is motor.HavenMotor
    motor_names = [m.name for m in motors]
    assert "SLT V Upper" in motor_names
    assert "SLT V Lower" in motor_names
    assert "SLT H Inbound" in motor_names
    # Check that the IOC name is set in labels
    motor1 = registry.find(name="SLT V Upper")
    assert "VME_crate" in motor1._ophyd_labels_


def test_motor_signals():
    m = motor.HavenMotor("motor_ioc", name="test_motor")
    assert m.description.pvname == "motor_ioc.DESC"
    assert m.tweak_value.pvname == "motor_ioc.TWV"
    assert m.tweak_forward.pvname == "motor_ioc.TWF"
    assert m.tweak_reverse.pvname == "motor_ioc.TWR"
    assert m.soft_limit_violation.pvname == "motor_ioc.LVIO"



from typing import Sequence, Optional
import warnings
import logging

from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky import plan_stubs as bps
import pandas as pd
from lmfit.model import Model
from lmfit.models import QuadraticModel

from haven.typing import Motor
from ..instrument.instrument_registry import registry
from .set_energy import set_energy
from .align_motor import align_pitch2, align_motor


__all__ = ["mono_ID_calibration"]


log = logging.getLogger(__name__)


def mono_ID_calibration(
    energies: Sequence,
    mono_motor: Motor = "energy_mono_energy",
    id_motor: Motor = "energy_id_energy",
    energy_motor: Motor = "energy",
    detector="I0",
    fit_model: Optional[Model] = None,
):
    """A bluesky plan to measure the offset between monochromator and
    insertion device energies.

    This plan will print the results of the equation. The fit results
    will also be added to the *fit_model* object as
    `fit_model.result`.

    Pseudo-code:

    - for energy in *energies*

      - Move ID to *energy*
      - Align pitch motor
      - 1D scan over mono energy motor
      - Calculate mono energy motor position for maximum flux

    - Fit ID energy vs mono energy with function

    Parameters
    ==========
    energies
      The target energies, in eV, over which the calibration will be
      performed.
    mono_motor
      The motor to move for monochromator energy (will be scanned).
    id_motor
      The motor to move for the insertion device (will not be
      scanned).
    energy_motor
      The motor to move when initially setting the energy for each
      point.
    detector
      The detector to measure.
    fit_model
      A lmfit model object that with determine the final fitting
      parameters for the calibration curve.

    """
    # Resolve device names
    mono_motor = registry.find(mono_motor)
    id_motor = registry.find(id_motor)
    results_df = pd.DataFrame(columns=["id_energy", "mono_energy"])
    # Do each energy point
    for energy in energies:
        # Set the mono and ID to the target energy to get started
        yield from set_energy(energy, positioners=[energy_motor])
        # Align the pitch motor
        yield from align_pitch2(detector=detector)
        # Determine maximum mono energy
        bec = BestEffortCallback()
        bec.disable_plots()
        bec.disable_table()
        detector_ = registry.find(detector)
        yield from align_motor(
            motor=id_motor, detector=detector_, distance=0.4, bec=bec
        )
        # Save the results to the dataframe for fitting later
        peak_center = bec.peaks["cen"]
        signal_name = getattr(detector_, "raw_counts", detector_).name
        try:
            mono_energy = mono_motor.user_readback.get()
        except AttributeError:
            mono_energy = mono_motor.readback.get()
        if signal_name in peak_center.keys():
            # results_df = results_df.append(
            new_row = {
                "id_energy": bec.peaks["cen"][signal_name],
                "mono_energy": mono_energy,
            }
            results_df = pd.concat(
                [results_df, pd.DataFrame([new_row])], ignore_index=True
            )
        print(results_df)
    # Fit the overall mono-ID calibration curve
    if fit_model is None:
        fit_model = QuadraticModel(nan_policy="omit")
    y = results_df.id_energy
    x = results_df.mono_energy
    params = fit_model.guess(y, x=x)
    try:
        fit = fit_model.fit(y, params, x=x)
    except Exception as e:
        log.warning(str(e))
        warnings.warn(str(e))
        print(fit_model, results_df)
    else:
        fit_model.fit_result = fit
        print(fit_model, fit, results_df)
    fit_model.results_df = results_df



import logging
import datetime as dt
from typing import Sequence
import yaml
from httpx import HTTPStatusError
from contextlib import contextmanager
from itertools import count

import numpy as np
from qtpy.QtWidgets import QWidget
from qtpy.QtGui import QStandardItemModel, QStandardItem
from qtpy.QtCore import Signal, Slot, QThread, Qt
from pyqtgraph import PlotItem, GraphicsLayoutWidget, PlotWidget, PlotDataItem
import qtawesome as qta
from matplotlib.colors import TABLEAU_COLORS

from firefly import display, FireflyApplication
from firefly.run_client import DatabaseWorker
from haven import tiled_client, load_config, exceptions

log = logging.getLogger(__name__)


colors = list(TABLEAU_COLORS.values())


class FiltersWidget(QWidget):
    returnPressed = Signal()

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        # Check for return keys pressed
        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            self.returnPressed.emit()


class Browser1DPlotItem(PlotItem):
    hover_coords_changed = Signal(str)

    def hoverEvent(self, event):
        super().hoverEvent(event)
        if event.isExit():
            self.hover_coords_changed.emit("NaN")
            return
        # Get data coordinates from event
        pos = event.scenePos()
        data_pos = self.vb.mapSceneToView(pos)
        pos_str = f"({data_pos.x():.3f}, {data_pos.y():.3f})"
        self.hover_coords_changed.emit(pos_str)


class Browser1DPlotWidget(PlotWidget):
    def __init__(self, parent=None, background="default", plotItem=None, **kargs):
        plot_item = Browser1DPlotItem(**kargs)
        super().__init__(parent=parent, background=background, plotItem=plot_item)


class RunBrowserDisplay(display.FireflyDisplay):
    runs_model: QStandardItemModel
    _run_col_names: Sequence = [
        "Plan",
        "Sample",
        "Edge",
        "E0",
        "Exit Status",
        "Datetime",
        "UID",
        "Proposal",
        "ESAF",
        "ESAF Users",
    ]
    _multiplot_items = {}

    # Signals
    runs_selected = Signal(list)
    runs_model_changed = Signal(QStandardItemModel)
    plot_1d_changed = Signal(object)
    filters_changed = Signal(dict)

    def __init__(self, root_node=None, args=None, macros=None, **kwargs):
        # self.prepare_run_client(root_node=root_node)
        super().__init__(args=args, macros=macros, **kwargs)
        self.start_run_client(root_node=root_node)

    def start_run_client(self, root_node):
        """Set up the database client in a separate thread."""
        # Create the thread and worker
        thread = QThread(parent=self)
        self._thread = thread
        worker = DatabaseWorker(root_node=root_node)
        self._db_worker = worker
        worker.moveToThread(thread)
        # Connect signals/slots
        thread.started.connect(worker.load_all_runs)
        worker.all_runs_changed.connect(self.set_runs_model_items)
        worker.selected_runs_changed.connect(self.update_metadata)
        worker.selected_runs_changed.connect(self.update_1d_signals)
        worker.selected_runs_changed.connect(self.update_1d_plot)
        worker.selected_runs_changed.connect(self.update_multi_plot)
        worker.db_op_started.connect(self.disable_run_widgets)
        worker.db_op_ended.connect(self.enable_run_widgets)
        self.runs_selected.connect(worker.load_selected_runs)
        self.ui.refresh_runs_button.clicked.connect(worker.load_all_runs)
        self.filters_changed.connect(worker.set_filters)
        worker.new_message.connect(self.show_message)
        # Make sure filters are current
        self.update_filters()
        # Start the thread
        thread.start()
        # Get distinct fields so we can populate the comboboxes
        worker.distinct_fields_changed.connect(self.update_combobox_items)
        worker.load_distinct_fields()

    def update_combobox_items(self, fields):
        for field_name, cb in [
            ("proposal_users", self.ui.filter_proposal_combobox),
            ("proposal_id", self.ui.filter_user_combobox),
            ("esaf_id", self.ui.filter_esaf_combobox),
            ("sample_name", self.ui.filter_sample_combobox),
            ("plan_name", self.ui.filter_plan_combobox),
            ("edge", self.ui.filter_edge_combobox),
        ]:
            cb.clear()
            cb.addItems(fields[field_name])

    def customize_ui(self):
        self.load_models()
        # Setup controls for select which run to show
        self.ui.run_tableview.selectionModel().selectionChanged.connect(
            self.update_selected_runs
        )
        self.ui.signal_y_combobox.currentTextChanged.connect(self.update_1d_plot)
        self.ui.signal_x_combobox.currentTextChanged.connect(self.update_1d_plot)
        self.ui.signal_r_combobox.currentTextChanged.connect(self.update_1d_plot)
        self.ui.signal_r_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.logarithm_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.invert_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.gradient_checkbox.stateChanged.connect(self.update_1d_plot)
        self.ui.plot_1d_hints_checkbox.stateChanged.connect(self.update_1d_signals)
        self.ui.refresh_runs_button.setIcon(qta.icon("fa5s.sync"))
        # Respond to filter controls getting updated
        self.ui.filter_user_combobox.currentTextChanged.connect(self.update_filters)
        self.ui.filter_proposal_combobox.currentTextChanged.connect(self.update_filters)
        self.ui.filter_sample_combobox.currentTextChanged.connect(self.update_filters)
        self.ui.filter_exit_status_combobox.currentTextChanged.connect(
            self.update_filters
        )
        self.ui.filter_esaf_combobox.currentTextChanged.connect(self.update_filters)
        self.ui.filter_current_proposal_checkbox.stateChanged.connect(
            self.update_filters
        )
        self.ui.filter_current_esaf_checkbox.stateChanged.connect(self.update_filters)
        self.ui.filter_plan_combobox.currentTextChanged.connect(self.update_filters)
        self.ui.filter_full_text_lineedit.textChanged.connect(self.update_filters)
        self.ui.filter_edge_combobox.currentTextChanged.connect(self.update_filters)
        self.ui.filters_widget.returnPressed.connect(self.refresh_runs_button.click)
        # Set up 1D plotting widgets
        self.plot_1d_item = self.ui.plot_1d_view.getPlotItem()
        self.plot_1d_item.addLegend()
        self.plot_1d_item.hover_coords_changed.connect(
            self.ui.hover_coords_label.setText
        )

    def get_signals(self, run, hinted_only=False):
        if hinted_only:
            xsignals = run.metadata["start"]["hints"]["dimensions"][0][0]
            ysignals = []
            hints = run["primary"].metadata["descriptors"][0]["hints"]
            for device, dev_hints in hints.items():
                ysignals.extend(dev_hints["fields"])
        else:
            xsignals = ysignals = run["primary"]["data"].keys()
        return xsignals, ysignals

    def set_runs_model_items(self, runs):
        self.runs_model.clear()
        self.runs_model.setHorizontalHeaderLabels(self._run_col_names)
        for run in runs:
            items = [QStandardItem(val) for val in run.values()]
            self.ui.runs_model.appendRow(items)
        # Adjust the layout of the data table
        sort_col = self._run_col_names.index("Datetime")
        self.ui.run_tableview.sortByColumn(sort_col, Qt.DescendingOrder)
        self.ui.run_tableview.resizeColumnsToContents()
        # Let slots know that the model data have changed
        self.runs_model_changed.emit(self.ui.runs_model)
        self.runs_total_label.setText(str(self.ui.runs_model.rowCount()))

    def disable_run_widgets(self):
        self.show_message("Loading...")
        widgets = [
            self.ui.run_tableview,
            self.ui.refresh_runs_button,
            self.ui.detail_tabwidget,
            self.ui.runs_total_layout,
            self.ui.filters_widget,
        ]
        for widget in widgets:
            widget.setEnabled(False)
        self.disabled_widgets = widgets
        self.setCursor(Qt.WaitCursor)

    def enable_run_widgets(self, exceptions=[]):
        if any(exceptions):
            self.show_message(exceptions[0])
        else:
            self.show_message("Done", 5000)
        # Re-enable the widgets
        for widget in self.disabled_widgets:
            widget.setEnabled(True)
        self.setCursor(Qt.ArrowCursor)

    def update_1d_signals(self, *args):
        # Store old values for restoring later
        comboboxes = [
            self.ui.signal_x_combobox,
            self.ui.signal_y_combobox,
            self.ui.signal_r_combobox,
        ]
        old_values = [cb.currentText() for cb in comboboxes]
        # Determine valid list of columns to choose from
        xcols = set()
        ycols = set()
        runs = self._db_worker.selected_runs
        use_hints = self.ui.plot_1d_hints_checkbox.isChecked()
        for run in runs:
            try:
                _xcols, _ycols = self.get_signals(run, hinted_only=use_hints)
            except KeyError:
                continue
            else:
                xcols.update(_xcols)
                ycols.update(_ycols)
        # Update the UI with the list of controls
        xcols = sorted(list(set(xcols)))
        ycols = sorted(list(set(ycols)))
        self.multi_y_signals = ycols
        for cb in [self.ui.multi_signal_x_combobox, self.ui.signal_x_combobox]:
            cb.clear()
            cb.addItems(xcols)
        for cb in [
            self.ui.signal_y_combobox,
            self.ui.signal_r_combobox,
        ]:
            cb.clear()
            cb.addItems(ycols)
        # Restore previous values
        for val, cb in zip(old_values, comboboxes):
            cb.setCurrentText(val)

    def calculate_ydata(
        self,
        x_data,
        y_data,
        r_data,
        x_signal,
        y_signal,
        r_signal,
        use_reference=False,
        use_log=False,
        use_invert=False,
        use_grad=False,
    ):
        """Take raw y and reference data and calculate a new y_data signal."""
        # Apply transformations
        y = y_data
        y_string = f"[{y_signal}]"
        try:
            if use_reference:
                y = y / r_data
                y_string = f"{y_string}/[{r_signal}]"
            if use_log:
                y = np.log(y)
                y_string = f"ln({y_string})"
            if use_invert:
                y *= -1
                y_string = f"-{y_string}"
            if use_grad:
                y = np.gradient(y, x_data)
                y_string = f"d({y_string})/d[{r_signal}]"
        except TypeError:
            msg = f"Could not calculate transformation."
            log.warning(msg)
            raise exceptions.InvalidTransformation(msg)
        return y, y_string

    def load_run_data(self, run, x_signal, y_signal, r_signal, use_reference=True):
        if "" in [x_signal, y_signal] or (use_reference and r_signal == ""):
            log.debug(
                f"Empty signal name requested: x='{x_signal}', y='{y_signal}', r='{r_signal}'"
            )
            raise exceptions.EmptySignalName
        try:
            data = run["primary"]["data"].read()
            y_data = data[y_signal]
            x_data = data[x_signal]
            if use_reference:
                r_data = data[r_signal]
            else:
                r_data = 1
        except KeyError as e:
            # No data, so nothing to plot
            msg = f"Cannot find key {e} in {run}."
            log.warning(msg)
            raise exceptions.SignalNotFound(msg)
        return x_data, y_data, r_data

    def multiplot_items(self, n_cols: int = 3):
        view = self.ui.plot_multi_view
        item0 = None
        for idx in count():
            row = int(idx / n_cols)
            col = idx % n_cols
            # Make a new plot item if one doesn't exist
            if (row, col) not in self._multiplot_items:
                self._multiplot_items[(row, col)] = view.addPlot(row=row, col=col)
            new_item = self._multiplot_items[(row, col)]
            # Link the X-axes together
            if item0 is None:
                item0 = new_item
            else:
                new_item.setXLink(item0)
            # Resize the viewing area to fit the contents
            width = view.width()
            plot_width = width / n_cols
            # view.resize(int(width), int(plot_width * row))
            view.setFixedHeight(1200)
            yield new_item

    def update_multi_plot(self, *args):
        x_signal = self.ui.multi_signal_x_combobox.currentText()
        if x_signal == "":
            return
        y_signals = self.multi_y_signals
        all_signals = set((x_signal, *y_signals))
        view = self.ui.plot_multi_view
        view.clear()
        self._multiplot_items = {}
        n_cols = 3
        runs = self._db_worker.selected_runs
        for run in runs:
            # print(run.metadata['start']['uid'], run['primary']['data'])
            data = run["primary"]["data"].read(all_signals)
            try:
                xdata = data[x_signal]
            except KeyError:
                log.warning(f"Cannot plot x='{x_signal}' for {list(data.keys())}")
                continue
            for y_signal, plot_item in zip(y_signals, self.multiplot_items()):
                # Get data from the database
                try:
                    plot_item.plot(xdata, data[y_signal])
                except KeyError:
                    log.warning(f"Cannot plot y='{y_signal}' for {list(data.keys())}")
                    continue
                else:
                    log.debug(f"Plotted {y_signal} vs. {x_signal} for {data}")
                finally:
                    plot_item.setTitle(y_signal)

    def update_1d_plot(self, *args):
        self.plot_1d_item.clear()
        # Figure out which signals to plot
        y_signal = self.ui.signal_y_combobox.currentText()
        x_signal = self.ui.signal_x_combobox.currentText()
        use_reference = self.ui.signal_r_checkbox.isChecked()
        if use_reference:
            r_signal = self.ui.signal_r_combobox.currentText()
        else:
            r_signal = None
        use_log = self.ui.logarithm_checkbox.isChecked()
        use_invert = self.ui.invert_checkbox.isChecked()
        use_grad = self.ui.gradient_checkbox.isChecked()
        # Do the plotting for each run
        y_string = ""
        x_data = None
        for idx, run in enumerate(self._db_worker.selected_runs):
            # Load datasets from the database
            try:
                x_data, y_data, r_data = self.load_run_data(
                    run, x_signal, y_signal, r_signal, use_reference=use_reference
                )
            except exceptions.SignalNotFound as e:
                self.show_message(str(e), 0)
                continue
            except exceptions.EmptySignalName:
                continue
            # Screen out non-numeric data types
            try:
                np.isfinite(x_data)
                np.isfinite(y_data)
                np.isfinite(r_data)
            except TypeError as e:
                msg = str(e)
                log.warning(msg)
                self.show_message(msg)
                continue
            # Calculate plotting data
            try:
                y_data, y_string = self.calculate_ydata(
                    x_data,
                    y_data,
                    r_data,
                    x_signal,
                    y_signal,
                    r_signal,
                    use_reference=use_reference,
                    use_log=use_log,
                    use_invert=use_invert,
                    use_grad=use_grad,
                )
            except exceptions.InvalidTransformation as e:
                self.show_message(str(e))
                continue
            # Plot this run's data
            color = colors[idx % len(colors)]
            self.plot_1d_item.plot(
                x=x_data,
                y=y_data,
                pen=color,
                name=run.metadata["start"]["uid"],
                clear=False,
            )
        # Axis formatting
        self.plot_1d_item.setLabels(left=y_string, bottom=x_signal)
        if x_data is not None:
            self.plot_1d_item.addLine(
                x=np.median(x_data), movable=True, label="{value:.3f}"
            )
        self.plot_1d_changed.emit(self.plot_1d_item)

    def update_metadata(self, *args):
        """Render metadata for the runs into the metadata widget."""
        # Combine the metadata in a human-readable output
        text = ""
        runs = self._db_worker.selected_runs
        for run in runs:
            md_dict = dict(**run.metadata)
            text += yaml.dump(md_dict)
            text += f"\n\n{'=' * 20}\n\n"
        # Update the widget with the rendered metadata
        self.ui.metadata_textedit.document().setPlainText(text)

    def update_selected_runs(self, *args):
        """Get the current runs from the database and stash them."""
        # Get UID's from the selection
        col_idx = self._run_col_names.index("UID")
        indexes = self.ui.run_tableview.selectedIndexes()
        uids = [i.siblingAtColumn(col_idx).data() for i in indexes]
        self.runs_selected.emit(uids)

    def update_filters(self, *args):
        new_filters = {
            "proposal": self.ui.filter_proposal_combobox.currentText(),
            "esaf": self.ui.filter_esaf_combobox.currentText(),
            "sample": self.ui.filter_sample_combobox.currentText(),
            "exit_status": self.ui.filter_exit_status_combobox.currentText(),
            "use_current_proposal": bool(
                self.ui.filter_current_proposal_checkbox.checkState()
            ),
            "use_current_esaf": bool(self.ui.filter_current_esaf_checkbox.checkState()),
            "plan": self.ui.filter_plan_combobox.currentText(),
            "full_text": self.ui.filter_full_text_lineedit.text(),
            "edge": self.ui.filter_edge_combobox.currentText(),
            "user": self.ui.filter_user_combobox.currentText(),
        }
        null_values = ["", False]
        new_filters = {k: v for k, v in new_filters.items() if v not in null_values}
        self.filters_changed.emit(new_filters)

    def load_models(self):
        # Set up the model
        self.runs_model = QStandardItemModel()
        # Add the model to the UI element
        self.ui.run_tableview.setModel(self.runs_model)

    def ui_filename(self):
        return "run_browser.ui"



import os
from pathlib import Path
import importlib

from haven import _iconfig
from haven._iconfig import load_config, print_config_value


def test_default_values():
    config = load_config()
    assert "beamline" in config.keys()


def test_loading_a_file():
    test_file = Path(__file__).resolve().parent / "test_iconfig.toml"
    config = load_config(file_paths=(test_file,))
    assert config["beamline"]["pv_prefix"] == "spam"


def test_config_files_from_env():
    # Set the environmental variable with the path to a test TOML file
    test_file = Path(__file__).resolve().parent / "test_iconfig.toml"
    os.environ["HAVEN_CONFIG_FILES"] = str(test_file)
    # Load the configuration
    importlib.reload(_iconfig)
    config = _iconfig.load_config()
    # Check that the test file was loaded
    assert config["beamline"]["pv_prefix"] == "spam"


def test_merging_dicts():
    """Do the entries from multiple dictioneries merge properly?"""
    this_dir = Path(__file__).resolve().parent
    default_files = [
        this_dir.parent / "haven" / "iconfig_default.toml",
        this_dir / "iconfig_testing.toml",
    ]
    test_file = this_dir / "test_iconfig.toml"
    config = load_config(file_paths=(*default_files, test_file))
    assert "description" in config["camera"]["camA"].keys()


def test_haven_config_cli(capsys):
    """Test the function used as a CLI way to get config values."""
    print_config_value(["monochromator.ioc"])
    # Check stdout for config value
    captured = capsys.readouterr()
    assert captured.out == "mono_ioc\n"



import time
import warnings
import logging

from bluesky.preprocessors import subs_decorator, subs_wrapper
from bluesky.callbacks import best_effort
from bluesky import plan_stubs as bps
from apstools.plans.alignment import lineup

from ..instrument.instrument_registry import registry
from ..preprocessors import shutter_suspend_decorator

log = logging.getLogger(__name__)


__all__ = ["align_motor", "align_pitch2"]


# @shutter_suspend_decorator()
def align_pitch2(
    distance=200, reverse=False, detector="I0", bec=None, feature="cen", md={}
):
    """Tune the monochromator 2nd crystal pitch motor.

    Find and set the position of maximum intensity in the ion chamber
    "I0". The scanning range is relative to the current motor
    position, and will go *distance* above and below it. For example,
    if the current position is 1000, ``distance=200`` will scan from
    800 to 1200.

    Parameters
    ==========
    bec
      A bluesky best effort callback for finding the peak position.
    distance
      Relative distance to scan in either direction.
    reverse
      Whether the scan goes low-to-high (False) or high-to-low (True).
    detector
      Which detector name to use.
    feature
      Which feature of the peak to use for alignment.
    md
      Extra metadata to pass into the run engine.

    """
    md_ = dict(plan_name="align_pitch2")
    md_.update(md)
    # Get motors
    pitch2 = registry.find(name="monochromator_pitch2")
    # Prepare and run the plan
    yield from align_motor(
        detector=detector,
        motor=pitch2,
        distance=distance,
        reverse=reverse,
        bec=bec,
        feature=feature,
        md=md_,
    )


@shutter_suspend_decorator()
def align_motor(
    detector, motor, distance=200, reverse=False, bec=None, feature="cen", md={}
):
    """Center the given motor using the beam intensity.

    Find and set the position of maximum intensity in the ion chamber
    *detector*. The scanning range is relative to the current motor
    position, and will go *distance* above and below it. For example,
    if the current position is 1000, ``distance=200`` will scan from
    800 to 1200.

    Parameters
    ==========
    bec
      A bluesky best effort callback for finding the peak position.
    distance
      Relative distance to scan in either direction.
    reverse
      Whether the scan goes low-to-high (False) or high-to-low (True).
    feature
      Which feature of the peak to use for alignment.
    md
      Extra metadata to pass into the run engine.

    """
    md_ = dict(plan_name="align_motor")
    md_.update(md)
    # Set up the best effort callback
    if bec is None:
        bec = best_effort.BestEffortCallback()
        bec.disable_table()
    # Determine plan parameters
    start, end = (distance, -distance) if reverse else (-distance, distance)
    # Resolve motors and detectors
    motor = registry.find(motor)
    det = registry.find(detector)
    detectors = [det]
    if hasattr(det, "raw_counts"):
        detectors.insert(0, det.raw_counts)
    # from pprint import pprint
    # pprint(detectors)
    # print(motor)
    # print(start, end)
    # print(feature)
    # print(bec, md_)
    # assert False
    plan = lineup(
        detectors, motor, start, end, npts=40, feature=feature, bec=bec, md=md_
    )
    plan = subs_wrapper(plan, bec)
    yield from plan
    # Wait for the callback to catch up
    t0 = time.time()
    timeout = 5
    det_name = detectors[0].name
    while time.time() - t0 < timeout:
        new_value = bec.peaks[feature].get(det_name)
        if new_value is not None:
            if feature in ["max", "min"]:
                # Max and min also include the y-value
                new_value, y_value = new_value
            break
    # Set the motor to the new value (from below accounting for hysteresis)
    if new_value is None:
        # Didn't find a peak position
        msg = (
            f"No peak position found for {det_name},"
            f"motor '{motor.name}' will not be set."
        )
        log.error(msg)
        warnings.warn(msg)
    else:
        yield from bps.mv(motor, new_value - abs(end - start) / 2)
        yield from bps.mv(motor, new_value)



from qtpy import QtWidgets
import qtawesome as qta


class RevealButton(QtWidgets.QPushButton):
    closed_icon = qta.icon("fa5s.angle-up")
    open_icon = qta.icon("fa5s.angle-down")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set up signals
        self.toggled.connect(self.toggle_icon)
        self.toggle_icon(self.isChecked())

    def toggle_icon(self, checked):
        if checked:
            self.setIcon(self.open_icon)
        else:
            self.setIcon(self.closed_icon)



import logging
from pathlib import Path
import warnings

from pydm.main_window import PyDMMainWindow

# from qtpy.QtCore import Slot
from qtpy import QtCore, QtGui, QtWidgets
from pydm import data_plugins
from haven.instrument import motor
from haven.instrument import motor
from haven import load_config

log = logging.getLogger(__name__)


class FireflyMainWindow(PyDMMainWindow):
    hide_nav_bar: bool = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.customize_ui()
        self.export_actions()

    def open(self, *args, **kwargs):
        widget = super().open(*args, **kwargs)
        # Connect signals for showing message in the window's status bar
        try:
            widget.status_message_changed.connect(self.show_status)
        except AttributeError:
            msg = f"No status messages on window: {args}, {kwargs}. Possibly you're not using FireflyMainWindow?"
            log.warning(msg)
            warnings.warn(msg)
        return widget

    def closeEvent(self, event):
        super().closeEvent(event)
        # Delete the window so it's recreated next time it's opened
        self.deleteLater()

    def add_menu_action(self, action_name: str, text: str, menu: QtWidgets.QMenu):
        """Add a new QAction to a menubar menu.

        The action triggers when the menu item is activated.

        Parameters
        ==========
        action_name
          The name of the parameter to the save the action. Will be
          accessible via *action_name* on the window, and window.ui.
        text
          Human-readable text to show on the menu item.
        menu
          The QMenu object in which to put this menu item.

        Returns
        =======
        action
          A QAction for the menu item.

        """
        action = QtWidgets.QAction(self)
        action.setObjectName(action_name)
        action.setText(text)
        menu.addAction(action)
        # Save the action as object parameters
        setattr(self.ui, action_name, action)
        setattr(self, action_name, action)
        return action

    def customize_ui(self):
        # Add window icon
        root_dir = Path(__file__).parent.absolute()
        icon_path = root_dir / "splash.png"
        self.setWindowIcon(QtGui.QIcon(str(icon_path)))
        # Hide the nav bar
        if self.hide_nav_bar:
            self.toggle_nav_bar(False)
            self.ui.actionShow_Navigation_Bar.setChecked(False)
        # Setup menu
        self.ui.menuSetup = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuSetup.setObjectName("menuSetup")
        self.ui.menuSetup.setTitle("&Setup")
        self.ui.menubar.addAction(self.ui.menuSetup.menuAction())
        # XAFS scan window
        self.add_menu_action(
            action_name="actionShow_Log_Viewer", text="Logs", menu=self.ui.menuView
        )
        # Positioners menu
        self.ui.menuPositioners = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuPositioners.setObjectName("menuPositioners")
        self.ui.menuPositioners.setTitle("Positioners")
        self.ui.menubar.addAction(self.ui.menuPositioners.menuAction())
        # Sample viewer
        self.add_menu_action(
            action_name="actionShow_Sample_Viewer",
            text="Sample",
            menu=self.ui.menuPositioners,
        )
        # Motors sub-menu
        self.ui.menuMotors = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuMotors.setObjectName("menuMotors")
        self.ui.menuMotors.setTitle("Motors")
        self.ui.menuPositioners.addAction(self.ui.menuMotors.menuAction())
        # Scans menu
        self.ui.menuScans = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuScans.setObjectName("menuScans")
        self.ui.menuScans.setTitle("Scans")
        self.ui.menubar.addAction(self.ui.menuScans.menuAction())
        # XAFS scan window
        self.add_menu_action(
            action_name="actionShow_Xafs_Scan", text="XAFS Scan", menu=self.ui.menuScans
        )
        # Detectors menu
        self.ui.menuDetectors = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuDetectors.setObjectName("menuDetectors")
        self.ui.menuDetectors.setTitle("Detectors")
        self.ui.menubar.addAction(self.ui.menuDetectors.menuAction())
        # Voltmeters window
        self.add_menu_action(
            action_name="actionShow_Voltmeters",
            text="Ion Chambers",
            menu=self.ui.menuDetectors,
        )
        # Cameras sub-menu
        self.ui.menuCameras = QtWidgets.QMenu(self.ui.menubar)
        self.ui.menuCameras.setObjectName("menuCameras")
        self.ui.menuCameras.setTitle("Cameras")
        self.ui.menuDetectors.addAction(self.ui.menuCameras.menuAction())
        # Add actions to the motors sub-menus
        from .application import FireflyApplication

        app = FireflyApplication.instance()
        for action in app.motor_actions:
            self.ui.menuMotors.addAction(action)
        # Add actions to the cameras sub-menus
        self.ui.menuCameras.addAction(app.show_cameras_window_action)
        self.ui.menuCameras.addSeparator()
        for action in app.camera_actions.values():
            self.ui.menuCameras.addAction(action)
        # Add area detectors to detectors menu
        ad_actions = app.area_detector_actions.values()
        if len(ad_actions) > 0:
            self.ui.menuDetectors.addSeparator()
        for action in ad_actions:
            self.ui.menuDetectors.addAction(action)
        # Add XRF detectors to detectors menu
        xrf_actions = app.xrf_detector_actions.values()
        if len(xrf_actions) > 0:
            self.ui.menuDetectors.addSeparator()
        for action in xrf_actions:
            self.ui.menuDetectors.addAction(action)
        # Add other menu actions
        self.ui.menuView.addAction(app.show_status_window_action)
        self.ui.menuView.addAction(app.launch_queuemonitor_action)
        self.ui.menuSetup.addAction(app.show_bss_window_action)
        self.ui.menuPositioners.addAction(app.show_energy_window_action)
        self.ui.menuScans.addSeparator()
        self.ui.menuScans.addAction(app.show_run_browser_action)

    def show_status(self, message, timeout=0):
        """Show a message in the status bar."""
        bar = self.statusBar()
        bar.showMessage(message, timeout)

    def update_window_title(self):
        if self.showing_file_path_in_title_bar:
            title = self.current_file()
        else:
            title = self.display_widget().windowTitle()
        # Add the beamline name
        config = load_config()
        beamline_name = config["beamline"]["name"]
        title += f" - {beamline_name} - Firefly"
        if data_plugins.is_read_only():
            title += " [Read Only Mode]"
        self.setWindowTitle(title)

    def export_actions(self):
        """Expose specific signals that might be useful for responding to window changes."""
        self.actionShow_Log_Viewer = self.ui.actionShow_Log_Viewer
        self.actionShow_Xafs_Scan = self.ui.actionShow_Xafs_Scan
        self.actionShow_Voltmeters = self.ui.actionShow_Voltmeters


class PlanMainWindow(FireflyMainWindow):
    """A Qt window that has extra controls for a bluesky runengine."""

    hide_nav_bar: bool = True

    def setup_navbar(self):
        # Remove previous navbar actions
        navbar = self.ui.navbar
        for action in navbar.actions():
            navbar.removeAction(action)
        # Add applications runengine actions
        from .application import FireflyApplication

        app = FireflyApplication.instance()
        navbar.addAction(app.start_queue_action)
        navbar.addSeparator()
        navbar.addAction(app.pause_runengine_action)
        navbar.addAction(app.pause_runengine_now_action)
        navbar.addSeparator()
        navbar.addAction(app.resume_runengine_action)
        navbar.addAction(app.stop_runengine_action)
        navbar.addAction(app.abort_runengine_action)
        navbar.addAction(app.halt_runengine_action)

    def customize_ui(self):
        super().customize_ui()
        self.setup_navbar()
        # Connect signals/slots
        from .application import FireflyApplication

        app = FireflyApplication.instance()
        app.queue_length_changed.connect(self.set_navbar_visibility)

    @QtCore.Slot(int)
    def set_navbar_visibility(self, queue_length: int):
        """Determine whether to make the navbar be visible."""
        log.debug(f"Setting navbar visibility. Queue length: {queue_length}")
        navbar = self.ui.navbar
        navbar.setVisible(queue_length > 0)



import logging
import warnings
from typing import Optional, Sequence

from ophyd import (
    CamBase,
    SingleTrigger,
    Kind,
    ADComponent as ADCpt,
    EpicsSignal,
)
from ophyd.areadetector.plugins import (
    ImagePlugin_V34,
    PvaPlugin_V34,
    OverlayPlugin,
    ROIPlugin_V34,
)


from .instrument_registry import registry
from .area_detector import DetectorBase, StatsPlugin_V34, SimDetector
from .._iconfig import load_config
from .. import exceptions


log = logging.getLogger(__name__)


__all__ = ["AravisDetector", "load_cameras"]


class AravisCam(CamBase):
    gain_auto = ADCpt(EpicsSignal, "GainAuto")
    acquire_time_auto = ADCpt(EpicsSignal, "ExposureAuto")


class AravisDetector(SingleTrigger, DetectorBase):
    """
    A gige-vision camera described by EPICS.
    """

    cam = ADCpt(AravisCam, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    overlays = ADCpt(OverlayPlugin, "Over1:")
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)


def load_cameras(config=None) -> Sequence[DetectorBase]:
    """Load cameras from config files and add to the registry.

    Returns
    =======
    Sequence[Camera]
      Sequence of the newly created and registered camera objects.

    """
    if config is None:
        config = load_config()
    print(config["camera"])
    # Get configuration details for the cameras
    devices = {
        k: v
        for (k, v) in config["camera"].items()
        if hasattr(v, "keys") and "prefix" in v.keys()
    }
    # Load each camera
    cameras = []
    for key, cam_config in devices.items():
        class_name = cam_config.get("device_class", "AravisDetector")
        camera_name = cam_config.get("name", key)
        description = cam_config.get("description", cam_config.get("name", key))
        DeviceClass = globals().get(class_name)
        # Check that it's a valid device class
        if DeviceClass is None:
            msg = f"camera.{key}.device_class={cam_config['device_class']}"
            raise exceptions.UnknownDeviceConfiguration(msg)
        try:
            device = DeviceClass(
                prefix=f"{cam_config['prefix']}:",
                name=camera_name,
                description=description,
                labels={"cameras"},
            )
        except TimeoutError as e:
            log.warning(f"Could not connect to camera: {camera_name}: {e}.")
        else:
            registry.register(device)
            cameras.append(device)
    return cameras



from ophyd import EpicsSignal, EpicsMotor
from pydm.data_plugins import plugin_for_address, add_plugin
from pydm import PyDMChannel

from firefly.ophyd_plugin import OphydConnection, OphydPlugin


def test_ophyd_pydm_ophyd_plugin(pydm_ophyd_plugin):
    plugin = plugin_for_address("oph://sim_detector")
    assert isinstance(plugin, OphydPlugin)


def test_signal_pv_lookup(sim_registry, pydm_ophyd_plugin):
    """Check that the device name gets converted to a PV for a simple ophyd.EpicsSignal."""
    # Create a ophyd signal
    signal = EpicsSignal("the_pv", name="epics_signal")
    sim_registry.register(signal)
    # Have the pydm_ophyd_plugin handle a channel
    channel = PyDMChannel(address="oph://epics_signal")
    pydm_ophyd_plugin.add_connection(channel)
    # Check that the connection was correctly retrieved from the
    # device registry
    connection = pydm_ophyd_plugin.connections["epics_signal"]
    assert connection.pv.pvname == "the_pv"


def test_device_pv_lookup(sim_registry, pydm_ophyd_plugin):
    """Check that the device name gets converted to a PV for a simple ophyd.EpicsSignal."""
    # Create an ophyd device
    signal = EpicsMotor("the_record", name="epics_motor")
    sim_registry.register(signal)
    # Have the pydm_ophyd_plugin handle a channel
    channel = PyDMChannel(address="oph://epics_motor_user_setpoint")
    pydm_ophyd_plugin.add_connection(channel)
    # Check that the connection was correctly retrieved from the
    # device registry
    connection = pydm_ophyd_plugin.connections["epics_motor_user_setpoint"]
    assert connection.pv.pvname == "the_record.VAL"


def test_bad_pv_lookup(sim_registry, pydm_ophyd_plugin):
    """Check that invalid or missing ophyd names are handled gracefully."""
    # Create a ophyd device
    signal = EpicsMotor("the_record", name="epics_motor")
    sim_registry.register(signal)
    # Look for a component that isn't a leaf (i.e. not an EpicsSignal)
    channel = PyDMChannel(address="oph://epics_motor")
    pydm_ophyd_plugin.add_connection(channel)
    # Check that the original address is set for widget feedback
    connection = pydm_ophyd_plugin.connections["epics_motor"]
    assert connection.pv.pvname == "epics_motor"
    # Look for a component that isn't in the registry
    channel = PyDMChannel(address="oph://jabberwocky")
    pydm_ophyd_plugin.add_connection(channel)
    # Check that the original address is set for widget feedback
    connection = pydm_ophyd_plugin.connections["jabberwocky"]
    assert connection.pv.pvname == "jabberwocky"



import time
import argparse
import cProfile
import logging
import pstats
import sys
from pathlib import Path


import haven


def main(default_fullscreen=False, default_display="status"):
    logger = logging.getLogger("")
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)-8s] - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel("INFO")
    handler.setLevel("INFO")

    try:
        """
        We must import QtWebEngineWidgets before creating a QApplication
        otherwise we get the following error if someone adds a WebView at Designer:
        ImportError: QtWebEngineWidgets must be imported before a QCoreApplication instance is created
        """
        from qtpy import QtWebEngineWidgets
    except ImportError:
        logger.debug("QtWebEngine is not supported.")

    from qtpy.QtWidgets import QSplashScreen
    from .application import FireflyApplication
    from qtpy.QtGui import QPixmap
    from qtpy import QtCore

    # Add plugins for handling ophyd objects
    from pydm.data_plugins import add_plugin
    from .ophyd_plugin import OphydPlugin

    add_plugin(OphydPlugin)

    # Set up splash screen
    fake_app = FireflyApplication(sys.argv)
    im_dir = Path(__file__).parent.resolve()
    im_fp = str(im_dir / "splash.png")
    pixmap = QPixmap(im_fp)
    splash = QSplashScreen(pixmap, QtCore.Qt.WindowStaysOnTopHint)
    splash.show()
    FireflyApplication.processEvents()
    for i in range(10):
        time.sleep(0.01)
    FireflyApplication.processEvents()

    from pydm import config

    # Set EPICS as the default protocol
    config.DEFAULT_PROTOCOL = "ca"

    ui_folder = Path(__file__).parent.resolve()

    from pydm.utilities import setup_renderer

    setup_renderer()

    import pydm
    from pydm.utilities.macro import parse_macro_string

    parser = argparse.ArgumentParser(description="Python Display Manager")
    parser.add_argument(
        "display",
        help="The name of a Firefly display to open.",
        nargs="?",
        default=default_display,
    )
    parser.add_argument(
        "--perfmon",
        action="store_true",
        help="Enable performance monitoring," + " and print CPU usage to the terminal.",
    )
    parser.add_argument(
        "--profile",
        action="store_true",
        help="Enable cProfile function profiling, printing on exit.",
    )
    parser.add_argument(
        "--hide-menu-bar",
        action="store_true",
        help="Start PyDM with the menu bar hidden.",
    )
    parser.add_argument(
        "--hide-status-bar",
        action="store_true",
        help="Start PyDM with the status bar hidden.",
    )
    if default_fullscreen:
        parser.add_argument(
            "--no-fullscreen",
            dest="fullscreen",
            action="store_false",
            help="Start Firefly in normal (non-fullscreen) mode.",
        )
    else:
        parser.add_argument(
            "--fullscreen",
            action="store_true",
            help="Start Firefly in full screen mode.",
        )
    parser.add_argument(
        "--read-only", action="store_true", help="Start Firefly in a Read-Only mode."
    )
    parser.add_argument(
        "--log_level",
        help="Configure level of log display",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="Firefly {version}".format(version=haven.__version__),
        help="Show Firefly's version number and exit.",
    )
    parser.add_argument(
        "-m",
        "--macro",
        help="Specify macro replacements to use, in JSON object format."
        + "    Reminder: JSON requires double quotes for strings, "
        + "so you should wrap this whole argument in single quotes."
        + '  Example: -m \'{"sector": "LI25", "facility": "LCLS"}\''
        + "--or-- specify macro replacements as KEY=value pairs "
        + " using a comma as delimiter  If you want to uses spaces "
        + " after the delimiters or around the = signs, "
        + " wrap the entire set with quotes "
        + '  Example: -m "sector = LI25, facility=LCLS"',
    )
    parser.add_argument(
        "--stylesheet",
        help="Specify the full path to a CSS stylesheet file, which"
        + " can be used to customize the appearance of Firefly and"
        + " Qt widgets.",
        default=None,
    )
    parser.add_argument(
        "display_args",
        help="Arguments to be passed to the PyDM client application"
        + " (which is a QApplication subclass).",
        nargs=argparse.REMAINDER,
        default=None,
    )

    pydm_args = parser.parse_args()
    if pydm_args.profile:
        profile = cProfile.Profile()
        profile.enable()

    macros = None
    if pydm_args.macro is not None:
        macros = parse_macro_string(pydm_args.macro)

    if pydm_args.log_level:
        logger.setLevel(pydm_args.log_level)
        handler.setLevel(pydm_args.log_level)

    app = FireflyApplication(
        display=pydm_args.display,
        command_line_args=pydm_args.display_args,
        perfmon=pydm_args.perfmon,
        hide_menu_bar=pydm_args.hide_menu_bar,
        hide_status_bar=pydm_args.hide_status_bar,
        fullscreen=pydm_args.fullscreen,
        read_only=pydm_args.read_only,
        macros=macros,
        stylesheet_path=pydm_args.stylesheet,
    )

    # Define devices on the beamline (slow!)
    app.load_instrument()
    FireflyApplication.processEvents()

    # Show the first window
    first_window = list(app.windows.values())[0]
    splash.finish(first_window)

    exit_code = app.exec_()

    if pydm_args.profile:
        profile.disable()
        stats = pstats.Stats(
            profile,
            stream=sys.stdout,
        ).sort_stats(pstats.SortKey.CUMULATIVE)
        stats.print_stats()

    sys.exit(exit_code)


def cameras():
    main(default_fullscreen=True, default_display="cameras")


if __name__ == "__main__":
    main()



"""Tools for modifying plans and data streams as they are generated."""


from typing import Union, Sequence, Iterable
from collections import ChainMap
import pkg_resources
import os
import socket
import getpass
import warnings
import logging

from bluesky.preprocessors import (
    baseline_wrapper as bluesky_baseline_wrapper,
    finalize_wrapper,
    suspend_wrapper,
)
from bluesky.suspenders import SuspendBoolLow
from bluesky.utils import make_decorator, Msg
from bluesky.preprocessors import msg_mutator
import epics


from .instrument.instrument_registry import registry
from .exceptions import ComponentNotFound
from ._iconfig import load_config


log = logging.getLogger()


def baseline_wrapper(
    plan,
    devices: Union[Sequence, str] = [
        "motors",
        "power_supplies",
        "xray_sources",
        "APS",
        "baseline",
    ],
    name: str = "baseline",
):
    bluesky_baseline_wrapper.__doc__
    # Resolve devices
    devices = registry.findall(devices, allow_none=True)
    yield from bluesky_baseline_wrapper(plan=plan, devices=devices, name=name)


def get_version(pkg_name):
    return pkg_resources.get_distribution(pkg_name).version


VERSIONS = dict(
    apstools=get_version("apstools"),
    bluesky=get_version("bluesky"),
    databroker=get_version("databroker"),
    epics_ca=epics.__version__,
    epics=epics.__version__,
    haven=get_version("haven-spc"),
    h5py=get_version("h5py"),
    matplotlib=get_version("matplotlib"),
    numpy=get_version("numpy"),
    ophyd=get_version("ophyd"),
    pymongo=get_version("pymongo"),
)


def inject_haven_md_wrapper(plan):
    """Inject additional metadata into a run.

    This takes precedences over the original metadata dict in the event of
    overlapping keys, but it does not mutate the original metadata dict.
    (It uses ChainMap.)

    Parameters
    ----------
    plan : iterable or iterator
        a generator, list, or similar containing `Msg` objects

    """

    def _inject_md(msg):
        if msg.command != "open_run":
            return msg
        # Prepare the metadata dictionary
        config = load_config()
        md = {
            # Software versions
            "versions": VERSIONS,
            # Controls
            "EPICS_HOST_ARCH": os.environ.get("EPICS_HOST_ARCH"),
            "epics_libca": os.environ.get("PYEPICS_LIBCA"),
            "EPICS_CA_MAX_ARRAY_BYTES": os.environ.get("EPICS_CA_MAX_ARRAY_BYTES"),
            # Facility
            "beamline_id": config["beamline"]["name"],
            "facility_id": config["facility"]["name"],
            "xray_source": config["facility"]["xray_source"],
            # Computer
            "login_id": f"{getpass.getuser()}@{socket.gethostname()}",
            "pid": os.getpid(),
            # User supplied
            "sample_name": "",
            # Bluesky
            "parameters": "",
            "purpose": "",
        }
        # Get metadata from the beamline scheduling system (bss)
        try:
            bss = registry.find(name="bss")
        except ComponentNotFound:
            if config["beamline"]["is_connected"]:
                wmsg = "Could not find bss device, metadata may be missing."
                warnings.warn(wmsg)
                log.warning(wmsg)
            bss_md = None
        else:
            bss_md = bss.get()
            from pprint import pprint

            pprint(bss_md._asdict())
            # pprint({k: v for k, v in bss_md})
            md.update(
                {
                    "proposal_id": bss_md.proposal.proposal_id,
                    "proposal_title": bss_md.proposal.title,
                    "proposal_users": bss_md.proposal.user_last_names,
                    "proposal_user_badges": bss_md.proposal.user_badges,
                    "esaf_id": bss_md.esaf.esaf_id,
                    "esaf_title": bss_md.esaf.title,
                    "esaf_users": bss_md.esaf.user_last_names,
                    "esaf_user_badges": bss_md.esaf.user_badges,
                    "mail_in_flag": bss_md.proposal.mail_in_flag,
                    "proprietary_flag": bss_md.proposal.proprietary_flag,
                    "bss_aps_cycle": bss_md.esaf.aps_cycle,
                    "bss_beamline_name": bss_md.proposal.beamline_name,
                }
            )
        # Update the message
        msg = msg._replace(kwargs=ChainMap(msg.kwargs, md))
        return msg

    return (yield from msg_mutator(plan, _inject_md))


def shutter_suspend_wrapper(plan, shutter_signals=None):
    """
    Install suspenders to the RunEngine, and remove them at the end.

    Parameters
    ----------
    plan : iterable or iterator
        a generator, list, or similar containing `Msg` objects
    suspenders : suspender or list of suspenders
        Suspenders to use for the duration of the wrapper

    Yields
    ------
    msg : Msg
        messages from plan, with 'install_suspender' and 'remove_suspender'
        messages inserted and appended
    """
    if shutter_signals is None:
        shutters = registry.findall("shutters", allow_none=True)
        shutter_signals = [s.pss_state for s in shutters]
    # Create a suspender for each shutter
    suspenders = []
    for sig in shutter_signals:
        suspender = SuspendBoolLow(sig, sleep=3.0)
        suspenders.append(suspender)
    if not isinstance(suspenders, Iterable):
        suspenders = [suspenders]

    def _install():
        for susp in suspenders:
            yield Msg("install_suspender", None, susp)

    def _remove():
        for susp in suspenders:
            yield Msg("remove_suspender", None, susp)

    def _inner_plan():
        yield from _install()
        return (yield from plan)

    return (yield from finalize_wrapper(_inner_plan(), _remove()))


baseline_decorator = make_decorator(baseline_wrapper)
shutter_suspend_decorator = make_decorator(shutter_suspend_wrapper)
